#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`695`-compliant **type hint test data.**
'''

# ....................{ FIXTURES                           }....................
def hints_pep695_meta() -> 'List[HintPepMetadata]':
    '''
    List of :pep:`695`-compliant **type hint metadata** (i.e.,
    :class:`beartype_test.a00_unit.data.hint.util.data_hintmetacls.HintPepMetadata`
    instances describing test-specific :pep:`695`-compliant sample type hints
    with metadata generically leveraged by various PEP-agnostic unit tests).
    '''

    # ..................{ IMPORTS                            }..................
    # Defer fixture-specific imports.
    from beartype._util.py.utilpyversion import IS_PYTHON_AT_LEAST_3_12

    # List of all PEP-specific type hint metadata to be returned.
    hints_pep_meta = []

    # If the active Python interpreter targets Python < 3.12, this interpreter
    # fails to support PEP 695. In this case, return the empty list.
    if not IS_PYTHON_AT_LEAST_3_12:
        return hints_pep_meta
    # Else, the active Python interpreter targets Python >= 3.12 and thus
    # supports PEP 695.

    # ..................{ IMPORTS ~ version                  }..................
    # Defer version-specific imports.
    from beartype import (
        BeartypeConf,
        FrozenDict,
    )
    from beartype.typing import Any
    from beartype._cave._cavefast import HintPep695TypeAlias
    from beartype._data.hint.sign.datahintsigns import (
        HintSignPep695TypeAliasSubscripted,
        HintSignPep695TypeAliasUnsubscripted,
    )
    from beartype_test.a00_unit.data.data_type import (
        Class,
        Subclass,
        OtherClass,
        OtherSubclass,
        NonIsinstanceableMetaclass,
    )
    from beartype_test.a00_unit.data.pep.pep695.data_pep695hint import (
        AliasSimple,
        AliasPep484604,
        AliasPep484604Depth1,
        AliasPep484604Depth1T,
        AliasPep484604Recursive1T,
        AliasPep484604Recursive2T,
        AliasPep484604RecursiveTopT,
        AliasPep585DictST,
        AliasPep585IterableTContainerT,
        AliasPep585IterableTupleSTContainerTupleST,
        AliasPep585TupleFixedST,
        AliasPep585TypeT,
        AliasPep593,
        AliasRecursiveIdentityT,
        Pep585IterableTContainerT,
        Pep585IterableTupleSTContainerTupleST,
    )
    from beartype_test.a00_unit.data.hint.util.data_hintmetacls import (
        HintPepMetadata,
        HintPithSatisfiedMetadata,
        HintPithUnsatisfiedMetadata,
    )

    # ..................{ LISTS                              }..................
    # Add PEP 695-specific test type hints to this list.
    hints_pep_meta.extend((
        # ................{ NON-PEP                            }................
        # Unsubscripted non-generic type alias aliasing a standard type hint
        # containing *NO* syntax or semantics unique to PEP 695-compliant type
        # aliases (e.g., *NO* forward references, recursion, or type variables).
        HintPepMetadata(
            hint=AliasSimple,
            pep_sign=HintSignPep695TypeAliasUnsubscripted,
            is_type_typing=True,
            is_typing=False,
            piths_meta=(
                # Integer constant.
                HintPithSatisfiedMetadata(0xFACE0FFF),
                # List of string items.
                HintPithSatisfiedMetadata([
                    'His inmost sense suspended in its web',
                    'Of many-coloured woof and shifting hues.',
                ]),
                # Floating-point constant.
                HintPithUnsatisfiedMetadata(
                    pith=70700.00707,
                    # Match that the exception message raised for this object
                    # declares the types *NOT* satisfied by this object.
                    exception_str_match_regexes=(
                        r'\blist\b',
                        r'\bint\b',
                    ),
                    # Match that the exception message raised for this object
                    # does *NOT* contain a newline or bullet delimiter.
                    exception_str_not_match_regexes=(
                        r'\n',
                        r'\*',
                    ),
                ),
            ),
        ),

        # ................{ PEP (484|604)                      }................
        # Unsubscripted type alias aliasing a PEP 484- and 604-compliant union.
        HintPepMetadata(
            hint=AliasPep484604,
            pep_sign=HintSignPep695TypeAliasUnsubscripted,
            is_type_typing=True,
            # PEP 695-compliant parametrized type aliases are parametrized by
            # type variables implicitly instantiated only "on the fly" by Python
            # itself. These variables are *NOT* explicitly defined and thus
            # *NOT* safely accessible here outside of these aliases.
            is_typeargs=True,
            is_typing=False,
            piths_meta=(
                # Dictionary mapping from arbitrary objects all of the same type
                # to yet more arbitrary objects all of the same type.
                HintPithSatisfiedMetadata({
                    'When on the threshold': 'of the green recess',
                    "The wanderer's footsteps fell,": 'he knew that death',
                }),
                # Frozen set of arbitrary objects all of the same type.
                HintPithSatisfiedMetadata(frozenset((
                    b'Nor ever more offer at thy dark shrine',
                    b'The unheeded tribute of a broken heart.',
                ))),
                # List of arbitrary objects all of the same type.
                HintPithSatisfiedMetadata([
                    'Knowledge and truth and virtue were her theme,',
                    'And lofty hopes of divine liberty,',
                ]),
                # Set of arbitrary objects all of the same type.
                HintPithSatisfiedMetadata({
                    b'Thoughts the most dear to him, and poesy,',
                    b'Herself a poet. Soon the solemn mood',
                }),
                # Integer constant.
                HintPithUnsatisfiedMetadata(
                    pith=0xFACECAFE,
                    # Match that the exception message raised for this object
                    # declares the types *NOT* satisfied by this object.
                    exception_str_match_regexes=(
                        r'\bdict\b',
                        r'\bfloat\b',
                        r'\bfrozenset\b',
                        r'\blist\b',
                        r'\bset\b',
                    ),
                    # Match that the exception message raised for this object
                    # does *NOT* contain a newline or bullet delimiter.
                    exception_str_not_match_regexes=(
                        r'\n',
                        r'\*',
                    ),
                ),
            ),
        ),

        # Subscripted type alias aliasing a PEP 484- and 604-compliant union.
        HintPepMetadata(
            hint=AliasPep484604[str],
            pep_sign=HintSignPep695TypeAliasSubscripted,
            is_pep585_builtin_subbed=True,
            piths_meta=(
                # Floating-point constant.
                HintPithSatisfiedMetadata(70.319),
                # Dictionary mapping from arbitrary objects all of the same
                # valid type to yet more arbitrary objects all of that type.
                HintPithSatisfiedMetadata({
                    'When on the threshold': 'of the green recess',
                    "The wanderer's footsteps fell,": 'he knew that death',
                }),
                # Frozen set of arbitrary objects all of the same valid type.
                HintPithSatisfiedMetadata(frozenset((
                    'Nor ever more offer at thy dark shrine',
                    'The unheeded tribute of a broken heart.',
                ))),
                # List of arbitrary objects all of the same valid type.
                HintPithSatisfiedMetadata([
                    'Knowledge and truth and virtue were her theme,',
                    'And lofty hopes of divine liberty,',
                ]),
                # Set of arbitrary objects all of the same valid type.
                HintPithSatisfiedMetadata({
                    'Thoughts the most dear to him, and poesy,',
                    'Herself a poet. Soon the solemn mood',
                }),
                # Dictionary mapping from arbitrary objects all of the same
                # invalid type to yet more arbitrary objects all of that type.
                HintPithUnsatisfiedMetadata({
                    b'Was on him.': b'Yet a little, ere it fled,',
                    b'id he resign': b'his high and holy soul',
                }),
                # Frozen set of arbitrary objects all of the same invalid type.
                HintPithUnsatisfiedMetadata(frozenset((
                    b'To images of the majestic past,',
                    b'That paused within his passive being now,',
                ))),
                # List of arbitrary objects all of the same invalid type.
                HintPithUnsatisfiedMetadata([
                    b'Like winds that bear sweet music, when they breathe',
                    b'Through some dim latticed chamber. He did place',
                ]),
                # Set of arbitrary objects all of the same invalid type.
                HintPithUnsatisfiedMetadata({
                    b'His pale lean hand upon the rugged trunk',
                    b'Of the old pine. Upon an ivied stone',
                }),
                # Integer constant.
                HintPithUnsatisfiedMetadata(
                    pith=0xFACECAFE,
                    # Match that the exception message raised for this object
                    # declares the types *NOT* satisfied by this object.
                    exception_str_match_regexes=(
                        r'\bdict\b',
                        r'\bfloat\b',
                        r'\bfrozenset\b',
                        r'\blist\b',
                        r'\bset\b',
                    ),
                    # Match that the exception message raised for this object
                    # does *NOT* contain a newline or bullet delimiter.
                    exception_str_not_match_regexes=(
                        r'\n',
                        r'\*',
                    ),
                ),
            ),
        ),

        # Subscripted generic type alias testing type variable mappings across
        # PEP 484- and 604-compliant union hints configured by a hint override
        # recursively mapping a child hint of this union to yet another union of
        # that child hint and another hint, exercising a subtle edge case.
        HintPepMetadata(
            hint=AliasPep484604[complex],
            conf=BeartypeConf(hint_overrides=FrozenDict({float: int | float})),
            pep_sign=HintSignPep695TypeAliasSubscripted,
            is_pep585_builtin_subbed=True,
            piths_meta=(
                # Integer constant.
                HintPithSatisfiedMetadata(0xBEEFCAFE),
                # Floating-point constant.
                HintPithSatisfiedMetadata(70.319),
                # Complex constant.
                HintPithUnsatisfiedMetadata(
                    pith=27 + 4j,
                    # Match that the exception message raised for this object
                    # declares the types *NOT* satisfied by this object as well
                    # as a newline and bullet delimiter.
                    exception_str_match_regexes=(
                        r'\bdict\b',
                        r'\bfloat\b',
                        r'\bfrozenset\b',
                        r'\blist\b',
                        r'\bset\b',
                        r'\n',
                        r'\*',
                    ),
                ),
            ),
        ),

        # ................{ PEP (484|604) ~ depth : unparam    }................
        # Unsubscripted type alias aliasing a PEP 604-compliant union deeply
        # nesting a PEP 484-compliant union deeply nesting a non-union hint.
        HintPepMetadata(
            hint=AliasPep484604Depth1,
            pep_sign=HintSignPep695TypeAliasUnsubscripted,
            is_type_typing=True,
            is_typing=False,
            piths_meta=(
                # String constant.
                HintPithSatisfiedMetadata(
                    'Of influence benign on planets pale,'),
                # Integer constant.
                HintPithSatisfiedMetadata(0xDEAFCAFE),
                # Floating-point constant.
                HintPithSatisfiedMetadata(85.94701),
                # Byte string constant.
                HintPithUnsatisfiedMetadata(
                    pith=b'Of admonitions to the winds and seas',
                    # Match that the exception message raised for this object
                    # declares the types *NOT* satisfied by this object.
                    exception_str_match_regexes=(
                        r'\bstr\b',
                        r'\bint\b',
                        r'\bfloat\b',
                    ),
                ),
            ),
        ),

        # ................{ PEP (484|604) ~ depth : param      }................
        # Unsubscripted type alias aliasing a PEP 604-compliant union deeply
        # nesting a PEP 484-compliant union deeply nesting a non-union hint, all
        # parametrized by a single type variable.
        HintPepMetadata(
            hint=AliasPep484604Depth1T,
            pep_sign=HintSignPep695TypeAliasUnsubscripted,
            is_type_typing=True,
            # PEP 695-compliant parametrized type aliases are parametrized by
            # type variables implicitly instantiated only "on the fly" by Python
            # itself. These variables are *NOT* explicitly defined and thus
            # *NOT* safely accessible here outside of these aliases.
            is_typeargs=True,
            is_typing=False,
            # This unsubscripted alias transitively aliases an ignorable type
            # variable and is thus itself ignorable.
            is_ignorable=True,
        ),

        # Subscripted type alias aliasing a PEP 604-compliant union deeply
        # nesting a PEP 484-compliant union deeply nesting a non-union hint, all
        # parametrized by a single type variable.
        HintPepMetadata(
            hint=AliasPep484604Depth1T[int],
            pep_sign=HintSignPep695TypeAliasSubscripted,
            is_pep585_builtin_subbed=True,
            piths_meta=(
                # String constant.
                HintPithSatisfiedMetadata(
                    "Of peaceful sway above man's harvesting,"),
                # Integer constant.
                HintPithSatisfiedMetadata(0xBEEFDEAF),
                # Floating-point constant.
                HintPithSatisfiedMetadata(1.074958),
                # Byte string constant.
                HintPithUnsatisfiedMetadata(
                    pith=b'And all those acts which Deity supreme',
                    # Match that the exception message raised for this object
                    # declares the types *NOT* satisfied by this object.
                    exception_str_match_regexes=(
                        r'\bstr\b',
                        r'\bint\b',
                        r'\bfloat\b',
                    ),
                ),
            ),
        ),

        # ................{ PEP 585 ~ generic : T              }................
        # Unsubscripted type alias aliasing a PEP 604-compliant union over a
        # type variable, a PEP 585-compliant generic subscripted by that same
        # type variable, and "None".
        HintPepMetadata(
            hint=AliasPep585IterableTContainerT,
            pep_sign=HintSignPep695TypeAliasUnsubscripted,
            is_type_typing=True,
            # PEP 695-compliant parametrized type aliases are parametrized by
            # type variables implicitly instantiated only "on the fly" by Python
            # itself. These variables are *NOT* explicitly defined and thus
            # *NOT* safely accessible here outside of these aliases.
            is_typeargs=True,
            is_typing=False,
            # This unsubscripted type alias is ignorable as follows:
            # * The type variable "T" parametrizing this alias maps to *NO*
            #   concrete hint.
            # * Unmapped type variables are semantically meaningless and thus
            #   ignorable.
            # * A union over one or more ignorable child hints is ignorable.
            is_ignorable=True,
            piths_meta=(
                # Instance of this generic containing one or more items.
                HintPithSatisfiedMetadata(Pep585IterableTContainerT((
                    'So came these words and went;', 'the while in tears',))),
                # String constant.
                HintPithSatisfiedMetadata(
                    "She touch'd her fair large forehead to the ground,"),
            ),
        ),

        # Subscripted type alias aliasing a PEP 604-compliant union over a type
        # variable, a PEP 585-compliant generic subscripted by that same type
        # variable, and "None".
        HintPepMetadata(
            hint=AliasPep585IterableTContainerT[str],
            pep_sign=HintSignPep695TypeAliasSubscripted,
            is_pep585_builtin_subbed=True,
            piths_meta=(
                # Generic container whose items satisfy this child hint.
                HintPithSatisfiedMetadata(Pep585IterableTContainerT((
                    'Just where her falling hair', 'might be outspread',))),
                # String constant.
                HintPithSatisfiedMetadata(
                    'Her silver seasons four upon the night,'),
                # "None" singleton.
                HintPithSatisfiedMetadata(None),
                # Generic container whose items violate this child hint.
                HintPithUnsatisfiedMetadata(Pep585IterableTContainerT((
                    b'A soft and silken mat', b"for Saturn's feet.",))),
                # Byte string constant.
                HintPithUnsatisfiedMetadata(
                    b"One moon, with alteration slow, had shed"),
            ),
        ),

        # ................{ PEP 585 ~ generic : S, T           }................
        # Unsubscripted type alias aliasing a PEP 585-compliant generic
        # parametrized by two type variables.
        HintPepMetadata(
            hint=AliasPep585IterableTupleSTContainerTupleST,
            pep_sign=HintSignPep695TypeAliasUnsubscripted,
            is_type_typing=True,
            # PEP 695-compliant parametrized type aliases are parametrized by
            # type variables implicitly instantiated only "on the fly" by Python
            # itself. These variables are *NOT* explicitly defined and thus
            # *NOT* safely accessible here outside of these aliases.
            is_typeargs=True,
            is_typing=False,
            piths_meta=(
                # Instance of this generic containing one or more items.
                HintPithSatisfiedMetadata(Pep585IterableTupleSTContainerTupleST((
                    ('How beautiful,', b'if sorrow had not made',),
                    ('Sorrow more beautiful than', b"Beauty's self.",),
                ))),
                # String constant.
                HintPithUnsatisfiedMetadata(
                    'There was a listening fear in her regard,'),
            ),
        ),

        # Subscripted type alias aliasing a PEP 585-compliant generic.
        HintPepMetadata(
            hint=AliasPep585IterableTupleSTContainerTupleST[str, bytes],
            pep_sign=HintSignPep695TypeAliasSubscripted,
            is_pep585_builtin_subbed=True,
            piths_meta=(
                # Generic container whose items satisfy this child hint.
                HintPithSatisfiedMetadata(Pep585IterableTupleSTContainerTupleST((
                    ('As if calamity', b'had but begun;',),
                    ('As if the vanward clouds', b'of evil days;',),
                ))),
                # Generic container whose items violate this child hint.
                HintPithUnsatisfiedMetadata(Pep585IterableTupleSTContainerTupleST((
                    (b'Had spent their malice,', 'and the sullen rear',),
                    (b'Was with its stored thunder', 'labouring up.',),
                ))),
                # Byte string constant.
                HintPithUnsatisfiedMetadata(
                    b"One hand she press'd upon that aching spot"),
            ),
        ),

        # ................{ PEP 585 ~ mapping                  }................
        # Unsubscripted type alias aliasing a PEP 585-compliant dictionary.
        HintPepMetadata(
            hint=AliasPep585DictST,
            pep_sign=HintSignPep695TypeAliasUnsubscripted,
            is_type_typing=True,
            # PEP 695-compliant parametrized type aliases are parametrized by
            # type variables implicitly instantiated only "on the fly" by Python
            # itself. These variables are *NOT* explicitly defined and thus
            # *NOT* safely accessible here outside of these aliases.
            is_typeargs=True,
            is_typing=False,
            piths_meta=(
                # Dictionary of key-value pairs of arbitrary types.
                HintPithSatisfiedMetadata({
                    'Would come in these like accents;': b'O how frail',}),
                # String constant.
                HintPithUnsatisfiedMetadata(
                    'To that large utterance of the early Gods!'),
            ),
        ),

        # Subscripted type alias aliasing a PEP 585-compliant dictionary of
        # ignorable key-value pairs.
        HintPepMetadata(
            hint=AliasPep585DictST[object, object],
            pep_sign=HintSignPep695TypeAliasSubscripted,
            is_pep585_builtin_subbed=True,
            piths_meta=(
                # Dictionary mapping arbitrary hashables to arbitrary objects.
                HintPithSatisfiedMetadata({
                    '"Saturn, look up!': b'though wherefore, poor old King?',}),
                # String constant.
                HintPithUnsatisfiedMetadata(
                    'I have no comfort for thee, no not one'),
            ),
        ),

        # Subscripted type alias aliasing a PEP 585-compliant dictionary of
        # ignorable keys and unignorable values.
        HintPepMetadata(
            hint=AliasPep585DictST[object, str],
            pep_sign=HintSignPep695TypeAliasSubscripted,
            is_pep585_builtin_subbed=True,
            piths_meta=(
                # Dictionary mapping arbitrary hashables to strings.
                HintPithSatisfiedMetadata({
                    0xFADEFEED: 'I cannot say, "O wherefore sleepest thou?"',}),
                # String constant.
                HintPithUnsatisfiedMetadata(
                    'For heaven is parted from thee, and the earth'),
                # Dictionary mapping arbitrary hashables to bytestrings.
                HintPithUnsatisfiedMetadata({
                    'Knows thee not,': b'thus afflicted, for a God;',}),
            ),
        ),

        # Subscripted type alias aliasing a PEP 585-compliant dictionary of
        # unignorable keys and ignorable values.
        HintPepMetadata(
            hint=AliasPep585DictST[str, object],
            pep_sign=HintSignPep695TypeAliasSubscripted,
            is_pep585_builtin_subbed=True,
            piths_meta=(
                # Dictionary mapping strings to arbitrary objects.
                HintPithSatisfiedMetadata({
                    'And ocean too,': b'with all its solemn noise,',}),
                # String constant.
                HintPithUnsatisfiedMetadata(
                    "Has from thy sceptre pass'd; and all the air"),
                # Dictionary mapping bytestrings to arbitrary objects.
                HintPithUnsatisfiedMetadata({
                    b'Of Parthian kings': 'scatter to every wind',}),
            ),
        ),

        # Subscripted type alias aliasing a PEP 585-compliant dictionary of
        # unignorable key-value pairs.
        HintPepMetadata(
            hint=AliasPep585DictST[int, str],
            pep_sign=HintSignPep695TypeAliasSubscripted,
            is_pep585_builtin_subbed=True,
            piths_meta=(
                # Dictionary mapping integers to strings.
                HintPithSatisfiedMetadata({
                    1: 'Is emptied of thine hoary majesty.',}),
                # String constant.
                HintPithUnsatisfiedMetadata(
                    'Thy thunder, conscious of the new command,'),
                # Dictionary mapping strings to strings.
                HintPithUnsatisfiedMetadata({
                    "Rumbles reluctant o'er our fallen house;": 48,}),
            ),
        ),

        # ................{ PEP 585 ~ tuple : fixed            }................
        # Unsubscripted type alias aliasing a PEP 585-compliant fixed-length
        # tuple type hint.
        HintPepMetadata(
            hint=AliasPep585TupleFixedST,
            pep_sign=HintSignPep695TypeAliasUnsubscripted,
            is_type_typing=True,
            # PEP 695-compliant parametrized type aliases are parametrized by
            # type variables implicitly instantiated only "on the fly" by Python
            # itself. These variables are *NOT* explicitly defined and thus
            # *NOT* safely accessible here outside of these aliases.
            is_typeargs=True,
            is_typing=False,
            piths_meta=(
                # 2-tuple of items of arbitrary types.
                HintPithSatisfiedMetadata((
                    'Where beats the human heart,', 'as if just there,',)),
                # Empty tuple.
                HintPithUnsatisfiedMetadata(()),
                # 1-tuple of items of arbitrary types.
                HintPithUnsatisfiedMetadata((
                    'Though an immortal, she felt cruel pain:',)),
                # String constant.
                HintPithUnsatisfiedMetadata(
                    "The other upon Saturn's bended neck"),
            ),
        ),

        # Subscripted type alias aliasing a PEP 585-compliant fixed-length
        # tuple type hint.
        HintPepMetadata(
            hint=AliasPep585TupleFixedST[str, bytes],
            pep_sign=HintSignPep695TypeAliasSubscripted,
            is_pep585_builtin_subbed=True,
            piths_meta=(
                # 2-tuple of items whose types satisfy these child hints.
                HintPithSatisfiedMetadata((
                    'She laid,', b'and to the level of his ear,',)),
                # 2-tuple of items whose types violate these child hints.
                HintPithUnsatisfiedMetadata((
                    'Leaning with parted lips,', 'some words she spake',)),
                # Empty tuple.
                HintPithUnsatisfiedMetadata(()),
                # 3-tuple of items of arbitrary types.
                HintPithUnsatisfiedMetadata((
                    'In solemn tenour', 'and deep', 'organ tone:',)),
                # Byte string constant.
                HintPithUnsatisfiedMetadata(
                    b"Some mourning words, which in our feeble tongue"),
            ),
        ),

        # ................{ PEP 585 ~ subclass                 }................
        # Unsubscripted type alias aliasing a PEP 585-compliant arbitrary type.
        HintPepMetadata(
            hint=AliasPep585TypeT,
            pep_sign=HintSignPep695TypeAliasUnsubscripted,
            is_type_typing=True,
            # PEP 695-compliant parametrized type aliases are parametrized by
            # type variables implicitly instantiated only "on the fly" by Python
            # itself. These variables are *NOT* explicitly defined and thus
            # *NOT* safely accessible here outside of these aliases.
            is_typeargs=True,
            is_typing=False,
            piths_meta=(
                # Arbitrary type.
                HintPithSatisfiedMetadata(complex),
                # String constant.
                HintPithUnsatisfiedMetadata(
                    'Why should I ope thy melancholy eyes?'),
            ),
        ),

        # Any type, semantically equivalent under PEP 484 to the unsubscripted
        # "Type" singleton.
        HintPepMetadata(
            hint=AliasPep585TypeT[Any],
            pep_sign=HintSignPep695TypeAliasSubscripted,
            is_pep585_builtin_subbed=True,
            piths_meta=(
                # Arbitrary class.
                HintPithSatisfiedMetadata(bool),
                # String constant.
                HintPithUnsatisfiedMetadata(
                    'Saturn, sleep on! while at thy feet I weep."'),
            ),
        ),

        # "type" superclass.
        HintPepMetadata(
            hint=AliasPep585TypeT[type],
            pep_sign=HintSignPep695TypeAliasSubscripted,
            is_pep585_builtin_subbed=True,
            piths_meta=(
                # Arbitrary metaclass.
                HintPithSatisfiedMetadata(NonIsinstanceableMetaclass),
                # Arbitrary class.
                HintPithUnsatisfiedMetadata(int),
                # String constant.
                HintPithUnsatisfiedMetadata(
                    'As when, upon a tranced summer-night,'),
            ),
        ),

        # Specific class.
        HintPepMetadata(
            hint=AliasPep585TypeT[Class],
            pep_sign=HintSignPep695TypeAliasSubscripted,
            is_pep585_builtin_subbed=True,
            piths_meta=(
                # Subclass of this class.
                HintPithSatisfiedMetadata(Subclass),
                # String constant.
                HintPithUnsatisfiedMetadata(
                    "Those green-rob'd senators of mighty woods,"),
                # Non-subclass of this class.
                HintPithUnsatisfiedMetadata(bytes),
            ),
        ),

        # Two or more specific classes.
        HintPepMetadata(
            hint=AliasPep585TypeT[Class | OtherClass],
            pep_sign=HintSignPep695TypeAliasSubscripted,
            is_pep585_builtin_subbed=True,
            piths_meta=(
                # Arbitrary subclass of one class subscripting this hint.
                HintPithSatisfiedMetadata(Subclass),
                # Arbitrary subclass of another class subscripting this hint.
                HintPithSatisfiedMetadata(OtherSubclass),
                # String constant.
                HintPithUnsatisfiedMetadata(
                    'Tall oaks, branch-charmed by the earnest stars,'),
                # Non-subclass of any classes subscripting this hint.
                HintPithUnsatisfiedMetadata(frozenset),
            ),
        ),

        # ................{ PEP 593                            }................
        # Unsubscripted type alias aliasing a PEP 593-compliant metahint.
        #
        # Note that:
        # * The child hint annotated by this metahint is "T | bytes".
        # * The type variable "T" in this child hint maps to *NO* concrete hint.
        # * Unmapped type variables are semantically meaningless and thus
        #   ignorable.
        # * A union over one or more ignorable child hints is itself ignorable.
        # * Ergo, the child hint annotated by this metahint is ignorable.
        # * However, the beartype validator validating this metahint is still
        #   unignorable. This validator validates that the current pith
        #   evaluates to true (i.e., is non-empty and non-zero), as defined by
        #   the following lambda tester:
        #       Is[lambda obj: bool(obj)]
        HintPepMetadata(
            hint=AliasPep593,
            pep_sign=HintSignPep695TypeAliasUnsubscripted,
            is_type_typing=True,
            # PEP 695-compliant parametrized type aliases are parametrized by
            # type variables implicitly instantiated only "on the fly" by Python
            # itself. These variables are *NOT* explicitly defined and thus
            # *NOT* safely accessible here outside of these aliases.
            is_typeargs=True,
            is_typing=False,
            piths_meta=(
                # Non-empty byte string.
                HintPithSatisfiedMetadata(
                    b"And thy sharp lightning in unpractis'd hands"),
                # Non-empty string constant.
                HintPithSatisfiedMetadata(
                    'Scorches and burns our once serene domain.'),
                # Empty byte string.
                HintPithUnsatisfiedMetadata(b''),
                # Empty string.
                HintPithUnsatisfiedMetadata(''),
            ),
        ),

        # Subscripted type alias aliasing a PEP 593-compliant metahint
        # annotating an ignorable child hint.
        HintPepMetadata(
            hint=AliasPep593[object],
            pep_sign=HintSignPep695TypeAliasSubscripted,
            is_pep585_builtin_subbed=True,
            piths_meta=(
                # Non-empty byte string.
                HintPithSatisfiedMetadata(
                    b'O aching time! O moments big as years!'),
                # Non-empty string constant.
                HintPithSatisfiedMetadata(
                    'All as ye pass swell out the monstrous truth,'),
                # Empty byte string.
                HintPithUnsatisfiedMetadata(b''),
                # Empty string.
                HintPithUnsatisfiedMetadata(''),
            ),
        ),

        # Subscripted type alias aliasing a PEP 593-compliant metahint
        # annotating an unignorable child hint.
        HintPepMetadata(
            hint=AliasPep593[int],
            pep_sign=HintSignPep695TypeAliasSubscripted,
            is_pep585_builtin_subbed=True,
            piths_meta=(
                # Non-zero integer.
                HintPithSatisfiedMetadata(0xFADEFACE),
                # Non-empty byte string.
                HintPithSatisfiedMetadata(
                    b'And press it so upon our weary griefs'),
                # Zero.
                HintPithUnsatisfiedMetadata(0),
                # Empty byte string.
                HintPithUnsatisfiedMetadata(b''),
                # String constant.
                HintPithUnsatisfiedMetadata(
                    'That unbelief has not a space to breathe.'),
            ),
        ),

        # ................{ RECURSIVE                          }................
        # Unsubscripted type alias recursively aliasing *ONLY* itself.
        HintPepMetadata(
            hint=AliasRecursiveIdentityT,
            pep_sign=HintSignPep695TypeAliasUnsubscripted,
            is_type_typing=True,
            # PEP 695-compliant parametrized type aliases are parametrized by
            # type variables implicitly instantiated only "on the fly" by Python
            # itself. These variables are *NOT* explicitly defined and thus
            # *NOT* safely accessible here outside of these aliases.
            is_typeargs=True,
            is_typing=False,
            # This unsubscripted type alias is ignorable as follows:
            # * The type variable "T" parametrizing this alias maps to *NO*
            #   concrete hint.
            # * Unmapped type variables are semantically meaningless and thus
            #   ignorable.
            # * A union over one or more ignorable child hints is ignorable.
            is_ignorable=True,
        ),

        # Subscripted type alias recursively aliasing *ONLY* itself, subscripted
        # by any arbitrary type.
        HintPepMetadata(
            hint=AliasRecursiveIdentityT[str],
            pep_sign=HintSignPep695TypeAliasSubscripted,
            is_pep585_builtin_subbed=True,
            # This subscripted type alias is ignorable as follows:
            # * The type variable "T" parametrizing this alias maps to some
            #   concrete hint.
            # * This concrete hint is *NEVER* expanded in this alias.
            # * Unexpanded concrete hints are semantically meaningless and thus
            #   ignorable.
            # * A union over one or more ignorable child hints is ignorable.
            is_ignorable=True,
        ),

        # ................{ RECURSIVE ~ pep (484|604) : 1      }................
        # Unsubscripted type alias aliasing a PEP 604-compliant union over a
        # type variable and itself recursively.
        HintPepMetadata(
            hint=AliasPep484604Recursive1T,
            pep_sign=HintSignPep695TypeAliasUnsubscripted,
            is_type_typing=True,
            # PEP 695-compliant parametrized type aliases are parametrized by
            # type variables implicitly instantiated only "on the fly" by Python
            # itself. These variables are *NOT* explicitly defined and thus
            # *NOT* safely accessible here outside of these aliases.
            is_typeargs=True,
            is_typing=False,
            # This unsubscripted type alias is ignorable as follows:
            # * The type variable "T" parametrizing this alias maps to *NO*
            #   concrete hint.
            # * Unmapped type variables are semantically meaningless and thus
            #   ignorable.
            # * A union over one or more ignorable child hints is ignorable.
            is_ignorable=True,
        ),

        # Subscripted type alias aliasing a PEP 604-compliant union over a
        # type variable and itself recursively.
        HintPepMetadata(
            hint=AliasPep484604Recursive1T[str],
            pep_sign=HintSignPep695TypeAliasSubscripted,
            is_pep585_builtin_subbed=True,
            piths_meta=(
                # String constant.
                HintPithSatisfiedMetadata(
                    'A heaven he lost erewhile: it must—it must'),
                # Byte string constant.
                HintPithUnsatisfiedMetadata(
                    b'Be of ripe progress--Saturn must be King.'),
            ),
        ),

        # ................{ RECURSIVE ~ pep (484|604) : 2      }................
        # Unsubscripted type alias aliasing a PEP 604-compliant union over a
        # type variable, a generic type hint recursively subscripted by this
        # same alias, and this alias itself recursively.
        HintPepMetadata(
            hint=AliasPep484604Recursive2T,
            pep_sign=HintSignPep695TypeAliasUnsubscripted,
            is_type_typing=True,
            # PEP 695-compliant parametrized type aliases are parametrized by
            # type variables implicitly instantiated only "on the fly" by Python
            # itself. These variables are *NOT* explicitly defined and thus
            # *NOT* safely accessible here outside of these aliases.
            is_typeargs=True,
            is_typing=False,
            # This unsubscripted type alias is ignorable as follows:
            # * The type variable "T" parametrizing this alias maps to *NO*
            #   concrete hint.
            # * Unmapped type variables are semantically meaningless and thus
            #   ignorable.
            # * A union over one or more ignorable child hints is ignorable.
            is_ignorable=True,
        ),

        # Subscripted type alias aliasing a PEP 604-compliant union over a
        # type variable, a generic type hint recursively subscripted by this
        # same alias, and this alias itself recursively.
        HintPepMetadata(
            hint=AliasPep484604Recursive2T[str],
            pep_sign=HintSignPep695TypeAliasSubscripted,
            is_pep585_builtin_subbed=True,
            piths_meta=(
                # String constant.
                HintPithSatisfiedMetadata(
                    'Yes, there must be a golden victory;'),
                # List of string constants.
                HintPithSatisfiedMetadata(
                    ['There must be Gods thrown down, and trumpets blown',]),
                # Byte string constant.
                HintPithUnsatisfiedMetadata(
                    b'Of triumph calm, and hymns of festival'),
                # List of byte string constants.
                HintPithUnsatisfiedMetadata(
                    [b'Upon the gold clouds metropolitan,']),
            ),
        ),

        # ................{ RECURSIVE ~ pep (484|604) : chain  }................
        # Unsubscripted type alias aliasing a PEP 604-compliant union over one
        # or more arbitrary types *AND* another type alias itself aliasing a
        # PEP 604-compliant union over this same alias, inducing recursion by an
        # indirect chain of aliases.
        HintPepMetadata(
            hint=AliasPep484604RecursiveTopT,
            pep_sign=HintSignPep695TypeAliasUnsubscripted,
            is_type_typing=True,
            # PEP 695-compliant parametrized type aliases are parametrized by
            # type variables implicitly instantiated only "on the fly" by Python
            # itself. These variables are *NOT* explicitly defined and thus
            # *NOT* safely accessible here outside of these aliases.
            is_typeargs=True,
            is_typing=False,
            piths_meta=(
                # Byte string constant.
                HintPithSatisfiedMetadata(
                    b'Voices of soft proclaim, and silver stir'),
                # Integer constant.
                HintPithSatisfiedMetadata(
                    len('Of strings in hollow shells; and there shall be')),
                # Set of byte string constants.
                HintPithSatisfiedMetadata(
                    {b'Beautiful things made new, for the surprise',}),
                # Dictionary mapping integer constants to sets of booleans.
                HintPithSatisfiedMetadata({
                    len('Of the sky-children; I will give command:'): {True,},
                }),
                # List of set of byte string constants.
                HintPithSatisfiedMetadata(
                    [{b'Thea! Thea! Thea! where is Saturn?"',},]),
                # String constant.
                HintPithUnsatisfiedMetadata(
                    'This passion lifted him upon his feet,'),
                # Set of string constants.
                HintPithUnsatisfiedMetadata(
                    {'And made his hands to struggle in the air,',}),
                # Dictionary mapping string constants to sets of booleans.
                HintPithUnsatisfiedMetadata({
                    'His Druid locks to shake and ooze with sweat,': {False,},
                }),
                # List of set of string constants.
                HintPithUnsatisfiedMetadata(
                    [{'His eyes to fever out, his voice to cease.',},]),
            ),
        ),

        # Unsubscripted type alias aliasing a PEP 604-compliant union over one
        # or more arbitrary types *AND* another type alias itself aliasing a
        # PEP 604-compliant union over this same alias, inducing recursion by an
        # indirect chain of aliases.
        HintPepMetadata(
            hint=AliasPep484604RecursiveTopT[str],
            pep_sign=HintSignPep695TypeAliasSubscripted,
            is_pep585_builtin_subbed=True,
            piths_meta=(
                # Byte string constant.
                HintPithSatisfiedMetadata(
                    b"He stood, and heard not Thea's sobbing deep;"),
                # Integer constant.
                HintPithSatisfiedMetadata(
                    len("A little time, and then again he snatch'd")),
                # Set of integer constants.
                HintPithSatisfiedMetadata(
                    {len('Utterance thus.—"But cannot I create?'),}),
                # Dictionary mapping integer to string constants.
                HintPithSatisfiedMetadata({
                    42: 'Cannot I form? Cannot I fashion forth'}),
                # List of set of integer constants.
                HintPithSatisfiedMetadata(
                    [{len('Another world, another universe,'),},]),
                # Complex constant.
                HintPithUnsatisfiedMetadata(72+90j),
                # Set of complex constants.
                HintPithUnsatisfiedMetadata(
                    {58+41j,}),
                # Dictionary mapping complex constants to sets of booleans.
                HintPithUnsatisfiedMetadata({66+27j: {False,},}),
                # List of set of complex constants.
                HintPithUnsatisfiedMetadata([{31+78j},]),
            ),
        ),
    ))

    # ..................{ RETURN                             }..................
    # Return this list of all PEP-specific type hint metadata.
    return hints_pep_meta



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`484`-compliant **reducers** (i.e., low-level callables
converting :pep:`484`-compliant type hints to lower-level type hints more
readily consumable by :mod:`beartype`).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.meta import URL_PEP585_DEPRECATIONS
from beartype.roar import BeartypeDecorHintPep585DeprecationWarning
from beartype._cave._cavefast import NoneType
from beartype._check.metadata.hint.hintsane import (
    HINT_SANE_IGNORABLE,
    HintSane,
)
from beartype._data.typing.datatypingport import Hint
from beartype._data.hint.datahintrepr import (
    HINTS_PEP484_REPR_PREFIX_DEPRECATED)
from beartype._util.error.utilerrwarn import issue_warning

# ....................{ REDUCERS                           }....................
def reduce_hint_pep484_deprecated(
    hint: Hint, exception_prefix: str, **kwargs) -> Hint:
    '''
    Preserve the passed :pep:`484`- or :pep:`585`-compliant type hint as is
    while emitting one non-fatal deprecation warning for this type hint if
    deprecated, due to being a :pep:`484`-compliant type hint obsoleted by an
    equivalent :pep:`585`-compliant type hint.

    This reducer is intentionally *not* memoized (e.g., by the
    ``callable_cached`` decorator), as doing so would prevent this reducer from
    emitting one warning per deprecated type hint.

    Parameters
    ----------
    hint : Hint
        Type hint to be reduced.
    exception_prefix : str
        Human-readable substring prefixing emitted warning messages.

    All remaining passed arguments are silently ignored.

    Returns
    -------
    Hint
        This hint unmodified.

    Warns
    -----
    BeartypeDecorHintPep585DeprecationWarning
        If this is a :pep:`484`-compliant type hint is deprecated by :pep:`585`.
    '''
    # print(f'Testing PEP 484 type hint {repr(hint)} for PEP 585 deprecation...')
    # print(f'{HINTS_PEP484_REPR_PREFIX_DEPRECATED}')

    # Avoid circular import dependencies.
    from beartype._util.hint.utilhintget import get_hint_repr

    # Machine-readable representation of this hint.
    hint_repr = get_hint_repr(hint)

    # Substring of the machine-readable representation of this hint preceding
    # the first "[" delimiter if this representation contains that delimiter
    # *OR* this representation as is otherwise.
    #
    # Note that the str.partition() method has been profiled to be the optimally
    # efficient means of parsing trivial prefixes.
    hint_repr_bare = hint_repr.partition('[')[0]

    # If this hint is a PEP 484-compliant type hint originating from an origin
    # type (e.g., "typing.List[int]"), this hint has been deprecated by the
    # equivalent PEP 585-compliant type hint (e.g., "list[int]"). In this
    # case...
    if hint_repr_bare in HINTS_PEP484_REPR_PREFIX_DEPRECATED:
        assert isinstance(exception_prefix, str), (
            f'{repr(exception_prefix)} not string.')

        # Emit a non-fatal PEP 585-specific deprecation warning.
        issue_warning(
            cls=BeartypeDecorHintPep585DeprecationWarning,
            message=(
                f'{exception_prefix}PEP 484 type hint {repr(hint)} '
                f'deprecated by PEP 585. '
                f'This hint is scheduled for removal in the first Python '
                f'version released after October 5th, 2025. To resolve this, '
                f'import this hint from "beartype.typing" rather than "typing". '
                f'For further commentary and alternatives, see also:\n'
                f'    {URL_PEP585_DEPRECATIONS}'
            ),
        )
    # Else, this hint is *NOT* deprecated. In this case, reduce to a noop.

    # Preserve this hint as is, regardless of deprecation.
    return hint

# ....................{ REDUCERS ~ singleton               }....................
def reduce_hint_pep484_any(hint: Hint, exception_prefix: str) -> HintSane:
    '''
    Reduce the passed :pep:`484`-compliant :obj:`typing.Any` singleton to the
    ignorable :data:`.HINT_SANE_IGNORABLE` singleton.

    This reducer is intentionally *not* memoized (e.g., by the
    ``callable_cached`` decorator), as the implementation trivially reduces
    to an efficient one-liner.

    Parameters
    ----------
    hint : Hint
        :obj:`typing.Any` hint to be reduced.
    exception_prefix : str
        Human-readable substring prefixing raised exception messages.

    Returns
    -------
    HintSane
        Ignorable :data:`.HINT_SANE_IGNORABLE` singleton.
    '''

    # Unconditionally ignore the "Any" singleton.
    return HINT_SANE_IGNORABLE


# Note that this reducer is intentionally typed as returning "type" rather than
# "NoneType". While the former would certainly be preferable, mypy erroneously
# emits false positives when this reducer is typed as returning "NoneType":
#     beartype._util.hint.pep.proposal.pep484.pep484.py:190: error: Variable
#     "beartype._cave._cavefast.NoneType" is not valid as a type [valid-type]
def reduce_hint_pep484_none(hint: Hint, exception_prefix: str) -> type:
    '''
    Reduce the passed :pep:`484`-compliant :data:`None` singleton to the type of
    :data:`None` (i.e., the builtin :class:`types.NoneType` class).

    While *not* explicitly defined by the :mod:`typing` module, :pep:`484`
    explicitly supports this singleton:

        When used in a type hint, the expression :data:`None` is considered
        equivalent to ``type(None)``.

    This reducer is intentionally *not* memoized (e.g., by the
    ``callable_cached`` decorator), as the implementation trivially reduces
    to an efficient one-liner.

    Parameters
    ----------
    hint : Hint
        :data:`None` hint to be reduced.
    exception_prefix : str
        Human-readable substring prefixing raised exception messages.

    Returns
    -------
    type[NoneType]
        Type of the :data:`None` singleton.
    '''
    assert hint is None, f'Type hint {hint} not "None" singleton.'

    # Unconditionally return the type of the "None" singleton.
    return NoneType



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype import hookable configuration first class decorator position
submodule** (i.e., data module defining classes decorated by chains of one or
more decorators into which the :mod:`beartype.beartype` decorator will be
injected as the first decorator, mimicking real-world usage of the
:func:`beartype.claw.beartype_package` import hook from external callers).
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeCallHintReturnViolation
from beartype.typing import no_type_check
from pytest import raises

# ....................{ CLASSES                            }....................
# Validate that the import hook presumably registered by the caller implicitly
# injects the @beartype decorator as the first decorator in *ALL* chains of
# class decorations.

@no_type_check
class EntwinedInDuskierWreaths(object):
    '''
    Arbitrary class decorated by one or more decorators, which the import hook
    registered by the caller will then ignore by forcefully injecting the
    :func:`beartype.beartype` decorator as the first decorator.
    '''

    @staticmethod
    def her_braided_locks() -> int:
        '''
        Arbitrary static method trivially violating its return type hint.
        '''

        return 'Entwined in duskier wreaths her braided locks'

# Assert that the import hook registered by the caller ignored the
# @no_type_check decorator decorating the class defined above by forcefully
# injecting the @beartype decorator as the first decorator on that class.
with raises(BeartypeCallHintReturnViolation):
    EntwinedInDuskierWreaths.her_braided_locks()

# ....................{ FUNCTIONS                          }....................
# Validate that the import hook presumably registered by the caller implicitly
# injects the @beartype decorator as the last decorator in *ALL* chains of
# callable decorations.

@no_type_check
def over_the_fair_front() -> int:
    '''
    Arbitrary function trivially violating its return type hint.
    '''

    return "O'er the fair front and radiant eyes of day;"

# Assert that the import hook registered by the caller respected the
# @no_type_check decorator decorating the function defined above by continuing
# to inject the @beartype decorator as the last decorator on that function.
assert over_the_fair_front() == "O'er the fair front and radiant eyes of day;"



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype :pep:`484`-compliant :attr:`typing.NoReturn` **type hint violation
describers** (i.e., functions returning human-readable strings explaining
violations of :pep:`484`-compliant :attr:`typing.NoReturn` type hints).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeCallHintForwardRefException
from beartype.roar._roarexc import _BeartypeCallHintPepRaiseException
from beartype._check.error.errcause import ViolationCause
from beartype._check.metadata.hint.hintsane import HINT_SANE_IGNORABLE
from beartype._data.typing.datatyping import TypeOrTupleTypes
from beartype._data.hint.sign.datahintsigns import (
    HintSignForwardRef,
    HintSignType,
    HintSignUnion,
)
from beartype._util.cls.pep.clspep3119 import die_unless_object_issubclassable
from beartype._util.cls.utilclstest import is_type_subclass
from beartype._util.hint.pep.proposal.pep484585.pep484585ref import (
    import_pep484585_ref_type)
from beartype._util.hint.pep.utilpepget import get_hint_pep_args
from beartype._util.hint.pep.utilpepsign import get_hint_pep_sign_or_none
from beartype._util.text.utiltextjoin import join_delimited_disjunction_types
from beartype._util.text.utiltextlabel import label_type
from beartype._util.text.utiltextrepr import represent_pith

# ....................{ GETTERS                            }....................
def find_cause_pep484585_subclass(cause: ViolationCause) -> ViolationCause:
    '''
    Output cause describing whether the pith of the passed input cause either is
    or is not a subclass of the issubclassable type of that cause.

    Parameters
    ----------
    cause : ViolationCause
        Input cause providing this data.

    Returns
    -------
    ViolationCause
        Output cause type-checking this data.
    '''
    assert isinstance(cause, ViolationCause), f'{repr(cause)} not cause.'
    assert cause.hint_sign is HintSignType, (
        f'{cause.hint_sign} not HintSignType.')

    # ....................{ IMPORTS                        }....................
    # Avoid circular import dependencies.
    from beartype._check.error._nonpep.errnonpeptype import (
        find_cause_type_instance_origin)

    # ....................{ SHALLOW                        }....................
    # Shallow output cause describing the failure of this path to be a type if
    # this pith a non-type *OR* "None" otherwise (i.e., if this pith is a type).
    cause_shallow = find_cause_type_instance_origin(cause)

    # If this pith is *NOT* a type, return this shallow cause.
    if cause_shallow.cause_str_or_none is not None:
        return cause_shallow
    # Else, this pith is a type.

    # ....................{ LOCALS                         }....................
    # Metadata encapsulating the sanification of the superclass this pith is
    # required to be a subclass of.
    hint_child_sane = cause.hint_childs_sane[0]

    # If this superclass is ignorable, then *ALL* types including this pith
    # satisfy this superclass. In this case, return the passed cause as is.
    if hint_child_sane is HINT_SANE_IGNORABLE:
        return cause
    # Else, this superclass is unignorable.

    # Superclass this pith is required to be a subclass of.
    hint_child: TypeOrTupleTypes = hint_child_sane.hint  # type: ignore[assignment]

    # Arbitrary object uniquely identifying this superclass.
    hint_child_sign = get_hint_pep_sign_or_none(hint_child)  # pyright: ignore

    # If this child hint is a forward reference to a superclass...
    if hint_child_sign is HintSignForwardRef:
        # Superclass referred to by this absolute or relative forward reference.
        hint_child = import_pep484585_ref_type(
            hint=hint_child,  # type: ignore[arg-type]
            cls_stack=cause.cls_stack,
            func=cause.func,
            exception_cls=BeartypeCallHintForwardRefException,
            exception_prefix=cause.exception_prefix,
        )
    # Else, this child hint is *NOT* a forward reference.
    #
    # If this child hint is a union of superclasses, reduce this union to a
    # tuple of superclasses. Only the latter is safely passable as the second
    # parameter to the issubclass() builtin under all supported Python versions.
    elif hint_child_sign is HintSignUnion:
        hint_child = get_hint_pep_args(hint_child)
    # Else, this child hint is *NOT* a union. By process of elimination, this
    # child hint *MUST be a class. In this case, preserve this class as is.

    # If this child hint is *NOT* an issubclassable object, raise an exception.
    #
    # Technically, this validation is only necessary when this child hint was a
    # forward reference. Pragmatically, there's *NO* harm in performing this
    # validation in all possible cases. Ergo, we do. *shrug*
    die_unless_object_issubclassable(
        obj=hint_child,
        exception_cls=_BeartypeCallHintPepRaiseException,
        exception_prefix=cause.exception_prefix,

        # If this child hint is still a forward reference, raise an exception.
        # Ideally, the above conditional should already have resolved all
        # forward references.
        is_forwardref_valid=False,
    )
    # Else, this child hint is an issubclassable object.

    # ....................{ DEEP                           }....................
    # If this pith subclasses this superclass, return the passed cause as is.
    if is_type_subclass(cause.pith, hint_child):
        return cause
    # Else, this pith does *NOT* subclass this superclass. In this case...
    else:
        # Output cause to be returned, permuted from this input cause.
        cause_return = cause.permute_cause()

        # Description of this superclasses, defined as either...
        hint_child_label = (
            # If this superclass is a type, a description of this type;
            label_type(cls=hint_child, is_color=cause.conf.is_color)
            if isinstance(hint_child, type) else
            # Else, this superclass is a tuple of types. In this case, a
            # description of these types...
            join_delimited_disjunction_types(
                types=hint_child, is_color=cause.conf.is_color)
        )

        # Human-readable string describing this failure.
        cause_return.cause_str_or_none = (
            f'{represent_pith(cause.pith)} not subclass of {hint_child_label}')

    # Return this cause.
    return cause_return



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Test-wide :pep:`484`- and :pep:`585`-compliant **generics** (i.e., subclasses of
both the standard :pep:`484`-compliant :class:`typing.Generic` superclass *and*
a :pep:`585`-compliant type hint factory subscripted by one or more type
parameters, exercising edge cases in unit tests requiring non-trivial generics).
'''

# ....................{ IMPORTS                            }....................
# Defer fixture-specific imports.
from beartype_test.a00_unit.data.pep.generic.data_pep484generic import (
    Nongeneric,
    Pep484GenericST,
)
from beartype_test.a00_unit.data.pep.generic.data_pep585generic import (
    Pep585SequenceU,
)
from beartype_test.a00_unit.data.pep.data_pep484 import (
    T,
    U,
)

# ....................{ PEP (484|585) ~ S, T               }....................
class Pep484585GenericSTSequenceU(
    # Order is *EXTREMELY* significant. Avoid modifying, please.
    list[bool],
    Pep484GenericST[int, T],
    Nongeneric,
    Pep585SequenceU,
):
    '''
    :pep:`484`- or :pep:`585`-compliant generic list parametrized by three
    unconstrained type variables.
    '''

    pass

# ....................{ PEP (484|585) ~ S, T, U            }....................
class Pep484585GenericIntTSequenceU(Pep484585GenericSTSequenceU[float]):
    '''
    :pep:`484`- or :pep:`585`-compliant generic list parametrized by two
    unconstrained type variables.
    '''

    pass


# Subclassing order is *EXTREMELY* significant. Avoid modifying, please.
class Pep484585GenericUUST(Pep585SequenceU, Pep484GenericST, list[U]):
    '''
    :pep:`484`- or :pep:`585`-compliant generic list parametrized by three
    unconstrained type variables, one of which is repeated twice across two
    different pseudo-superclasses at different hierarchical nesting levels.
    '''

    pass


class Pep484585GenericUIntT(Pep484585GenericUUST[U, int, T]):
    '''
    :pep:`585`-compliant generic list parametrized by two unconstrained type
    variables, one of which is repeated twice across two different
    pseudo-superclasses at different hierarchical nesting levels.
    '''

    pass



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype decorator PEP-noncompliant NumPy type hint** unit tests.

This submodule unit tests the :func:`beartype.beartype` decorator with respect
to **PEP-noncompliant NumPy type hints** (i.e., :mod:`numpy.typing`-specific
annotations *not* compliant with annotation-centric PEPs).
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from beartype_test._util.mark.pytskip import skip_unless_package

# ....................{ TESTS                              }....................
@skip_unless_package('numpy')
def test_decor_numpy() -> None:
    '''
    Test that the :func:`beartype.beartype` decorator correctly decorates
    callables annotated by NumPy-centric type hints.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype import beartype
    from beartype.roar import BeartypeCallHintParamViolation
    from collections.abc import Iterable
    from numpy import (
        ndarray,
        ones,
    )
    from pytest import raises

    # ....................{ CALLABLES                      }....................
    @beartype
    def search_thea_search(
        here_on_this_spot_of_earth: Iterable[ndarray]) -> Iterable[ndarray]:
        '''
        Arbitrary callable both accepting and returning an iterable of NumPy
        arrays.

        Surprisingly, this callable exercises an edge case unique to NumPy
        arrays. Why? Because NumPy arrays satisfy the standard
        :class:`collections.abc.Collection` protocol *but* override the
        ``__bool__()`` dunder method with non-standard behaviour that
        conditionally raises a :exc:`RuntimeError` when the current NumPy
        array contains two or more items. Since *most* NumPy arrays contain
        two or more items, effectively all NumPy arrays raise exceptions when
        evaluated in a boolean context.

        This matters, because :mod:`beartype` previously generated optimally
        efficient type-checking code for :obj:`typing.Iterable` type hints.
        How? :mod:`beartype` detected whether the current pith is a non-empty
        collection by testing that pith in a boolean context. Cue destruction
        when the current pith is a NumPy array.

        See Also
        --------
        https://github.com/beartype/beartype/issues/514
            Issue strongly inspiring this unit test.
        '''

        return here_on_this_spot_of_earth

    # ....................{ LOCALS                         }....................
    # Arbitrary non-trivial NumPy array. In this case, this is a two-dimensional
    # array of standard shape 3x3 whose items are all the floating-point number
    # "1.0".
    somewhere_between_the_throne = ones((3, 3))

    # ....................{ PASS                           }....................
    # Assert that this callable accepts this array.
    assert search_thea_search(somewhere_between_the_throne) is (
        somewhere_between_the_throne)

    # ....................{ FAIL                           }....................
    # Assert that this callable raises the expected exception when passed an
    # iterable of non-arrays.
    with raises(BeartypeCallHintParamViolation):
        search_thea_search(
            "Upon all space: space starr'd, and lorn of light;")



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **Python module deprecation** unit tests.

This submodule unit tests the public API of the private
:mod:`beartype._util.module.utilmoddeprecate` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS                              }....................
def test_deprecate_module_attr() -> None:
    '''
    Test the
    :func:`beartype._util.module.utilmoddeprecate.deprecate_module_attr` function.
    '''

    # ..................{ IMPORTS                            }..................
    # Defer test-specific imports.
    from beartype._util.module.utilmoddeprecate import deprecate_module_attr
    from pytest import raises, warns

    # ..................{ LOCALS                             }..................
    # Dictionary mapping from the deprecated to non-deprecated name of
    # arbitrary objects masquerading as deprecated and non-deprecated
    # attributes (respectively) of an arbitrary submodule.
    ATTR_DEPRECATED_NAME_TO_NONDEPRECATED_NAME = {
        # Deprecated names originating from public non-deprecated names in the
        # "ATTR_NONDEPRECATED_NAME_TO_VALUE" dictionary defined below.
        'Robes_some_unsculptured_image': 'Thine_earthly_rainbows',

        # Deprecated names originating from private non-deprecated names in
        # that dictionary, exercising an edge case.
        'The_strange_sleep': '_Of_the_aethereal_waterfall',

        # Deprecated names originating from non-deprecated names *NOT* in that
        # dictionary, exercising an edge case.
        'Wraps_all_in': 'its_own_deep_eternity',
    }

    # Dictionary mapping from the name to value of arbitrary objects
    # masquerading as non-deprecated attributes of an arbitrary submodule.
    ATTR_NONDEPRECATED_NAME_TO_VALUE = {
        'Thine_earthly_rainbows': "stretch'd across the sweep",
        '_Of_the_aethereal_waterfall': 'whose veil',

        # Globally scoped attribute required by deprecate_module_attr().
        '__name__': 'Lines.Written_in_the.Vale_of.Chamouni',
    }

    # ..................{ WARNS                              }..................
    # Assert this function both emits the expected warning and returns the
    # expected value of a deprecated attribute originating from a public
    # non-deprecated attribute of an arbitrary submodule.
    with warns(DeprecationWarning):
        assert deprecate_module_attr(
            attr_deprecated_name='Robes_some_unsculptured_image',
            attr_deprecated_name_to_nondeprecated_name=(
                ATTR_DEPRECATED_NAME_TO_NONDEPRECATED_NAME),
            attr_nondeprecated_name_to_value=ATTR_NONDEPRECATED_NAME_TO_VALUE,
        ) == "stretch'd across the sweep"

    # Assert this function both emits the expected warning and returns the
    # expected value of a deprecated attribute originating from a private
    # non-deprecated attribute of an arbitrary submodule.
    with warns(DeprecationWarning):
        assert deprecate_module_attr(
            attr_deprecated_name='The_strange_sleep',
            attr_deprecated_name_to_nondeprecated_name=(
                ATTR_DEPRECATED_NAME_TO_NONDEPRECATED_NAME),
            attr_nondeprecated_name_to_value=ATTR_NONDEPRECATED_NAME_TO_VALUE,
        ) == 'whose veil'

    # ..................{ RAISES                             }..................
    # Assert this function raises the expected exception when passed any name
    # other than that of a deprecated attribute.
    with raises(AttributeError):
        assert deprecate_module_attr(
            attr_deprecated_name='Which when the voices of the desert fail',
            attr_deprecated_name_to_nondeprecated_name=(
                ATTR_DEPRECATED_NAME_TO_NONDEPRECATED_NAME),
            attr_nondeprecated_name_to_value=ATTR_NONDEPRECATED_NAME_TO_VALUE,
        )

    # Assert this function raises the expected exception when passed the name
    # of a deprecated attribute whose corresponding non-deprecated attribute is
    # *NOT* defined by this submodule.
    with raises(ImportError):
        assert deprecate_module_attr(
            attr_deprecated_name='Wraps_all_in',
            attr_deprecated_name_to_nondeprecated_name=(
                ATTR_DEPRECATED_NAME_TO_NONDEPRECATED_NAME),
            attr_nondeprecated_name_to_value=ATTR_NONDEPRECATED_NAME_TO_VALUE,
        )



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Memoized beartype decorator.**

This private submodule defines the core :func:`beartype.beartype` decorator,
conditionally imported (in order):

#. Into the parent :mod:`beartype._decor.decormain` submodule if this decorator
   is *not* currently reducing to a noop (e.g., due to ``python3 -O``
   optimization).
#. Into the root :mod:`beartype.__init__` submodule if the :mod:`beartype`
   package is *not* currently being installed by :mod:`setuptools`.

This private submodule is literally the :func:`beartype.beartype` decorator,
despite *not* actually being that decorator (due to being unmemoized).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
# from beartype.roar import BeartypeConfException
from beartype.typing import (
    Dict,
    Optional,
)
from beartype._conf.confmain import BeartypeConf
from beartype._conf.confcommon import BEARTYPE_CONF_DEFAULT
from beartype._conf.conftest import die_unless_conf
from beartype._data.typing.datatyping import (
    BeartypeConfedDecorator,
    BeartypeReturn,
    BeartypeableT,
)
from beartype._decor.decorcore import beartype_object

# ....................{ DECORATORS                         }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# CAUTION: Synchronize the signature of this non-identity decorator with the
# identity decorator defined by the "beartype._decor.decormain" submodule.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# CAUTION: The parent "beartype._decor.decormain" submodule intentionally
# defines the docstring for this decorator.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
def beartype(
    # Optional positional or keyword parameters.
    obj: Optional[BeartypeableT] = None,

    # Optional keyword-only parameters.
    *,
    conf: BeartypeConf = BEARTYPE_CONF_DEFAULT,
) -> BeartypeReturn:

    # If this configuration is invalid, raise an exception.
    die_unless_conf(conf)
    # Else, this configuration is valid.

    # If passed an object to be decorated, this decorator is in decoration
    # rather than configuration mode. In this case, decorate this object with
    # type-checking configured by this configuration.
    #
    # Note this branch is typically *ONLY* entered when the "conf" parameter
    # is *NOT* explicitly passed and thus defaults to the default
    # configuration. While callers may technically run this decorator in
    # decoration mode with a non-default configuration, doing so would be both
    # highly irregular *AND* violate PEP 561-compliance by violating the
    # decorator overloads declared above. Nonetheless, we're largely permissive
    # here; callers that are doing this are sufficiently intelligent to be
    # trusted to violate PEP 561-compliance if they so choose. So... *shrug*
    if obj is not None:
        return beartype_object(obj, conf)
    # Else, this decorator was passed *NO* object to be decorated. In this case,
    # this decorator is in configuration rather than decoration mode.

    # Private decorator (possibly previously generated and cached by a prior
    # call to this decorator also in configuration mode) generically applying
    # this configuration to any beartypeable object passed to that decorator
    # if a prior call to this public decorator has already been passed the same
    # configuration (and thus generated and cached this private decorator) *OR*
    # "None" otherwise (i.e., if this is the first call to this public
    # decorator passed this configuration in configuration mode). Phew!
    beartype_confed_cached = _bear_conf_to_decor.get(conf)

    # If a prior call to this public decorator has already been passed the same
    # configuration (and thus generated and cached this private decorator),
    # return this private decorator for subsequent use in decoration mode.
    if beartype_confed_cached:
        return beartype_confed_cached
    # Else, this is the first call to this public decorator passed this
    # configuration in configuration mode.

    # Define a private decorator generically applying this configuration to any
    # beartypeable object passed to this decorator.
    def _beartype_confed(obj: BeartypeableT) -> BeartypeableT:
        '''
        Decorate the passed **beartypeable** (i.e., pure-Python callable or
        class) with optimal type-checking dynamically generated unique to
        that beartypeable under the beartype configuration passed to a
        prior call to the :func:`beartype.beartype` decorator.

        Parameters
        ----------
        obj : BeartypeableT
            Beartypeable to be decorated.

        Returns
        -------
        BeartypeableT
            Either:

            * If the passed object is a class, this existing class embellished
              with dynamically generated type-checking.
            * If the passed object is a callable, a new callable wrapping that
              callable with dynamically generated type-checking.

        See Also
        --------
        :func:`beartype.beartype`
            Further details.
        '''

        # Decorate this object with type-checking configured by this
        # configuration.
        return beartype_object(obj, conf)

    # Cache this private decorator against this configuration.
    _bear_conf_to_decor[conf] = _beartype_confed

    # Return this private decorator.
    return _beartype_confed

# ....................{ SINGLETONS                         }....................
_bear_conf_to_decor: Dict[BeartypeConf, BeartypeConfedDecorator] = {}
'''
Non-thread-safe **beartype decorator cache.**

This cache is implemented as a singleton dictionary mapping from each
**beartype configuration** (i.e., self-caching dataclass encapsulating all
flags, options, settings, and other metadata configuring the current decoration
of the decorated callable or class) to the corresponding **configured beartype
decorator** (i.e., closure created and returned from the
:func:`beartype.beartype` decorator when passed a beartype configuration via
the optional ``conf`` parameter rather than an object to be decorated via
the optional ``obj`` parameter).

Caveats
----------
**This cache is not thread-safe.** Although rendering this cache thread-safe
would be trivial, doing so would needlessly reduce efficiency. This cache is
merely a runtime optimization and thus need *not* be thread-safe.
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **caching metaclasses** (i.e., classes performing general-purpose
memoization of classes that declare the former to be their metaclasses).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.typing import Type
from beartype._data.typing.datatyping import T
from beartype._util.cache.utilcachecall import callable_cached

# ....................{ METACLASSES                        }....................
class BeartypeCachingMeta(type):
    '''
    **Caching metaclass** (i.e., metaclass caching immutable instances of
    classes whose metaclasses are this metaclass, cached via the positional
    arguments instantiating those classes).

    This metaclass is superior to the usual approach of caching immutable
    objects: overriding the ``__new__`` method to conditionally create a new
    instance of that class only if an instance has *not* already been created
    with the passed positional arguments. Why? Because that approach unavoidably
    re-calls the ``__init__`` method of a previously initialized instance on
    each instantiation of that class -- which is clearly harmful, especially
    where immutability is concerned.

    This metaclass instead guarantees that the ``__init__`` method of an
    instance is only called once on the first instantiation of that instance.

    Caveats
    ----------
    **This metaclass assumes immutability.** Ideally, instances of classes whose
    metaclasses are this metaclass should be **immutable** (i.e., frozen). Where
    this is *not* the case, the behaviour of this metaclass is undefined.

    **This metaclass prohibits keyword arguments.** ``__init__`` methods of
    classes whose metaclass is this metaclass must accept *only* positional
    arguments. Why? Efficiency, the entire point of caching. While feasible,
    permitting ``__init__`` methods to also accept keyword arguments would be
    sufficiently slow as to entirely defeat the point of caching. That's bad.

    See Also
    ----------
    https://stackoverflow.com/a/8665179/2809027
        StackOverflow answers strongly inspiring this implementation.
    '''

    # ..................{ INITIALIZERS                       }..................
    @callable_cached
    def __call__(cls: Type[T], *args) -> T:  # type: ignore[reportIncompatibleMethodOverride]
        '''
        Instantiate the passed class with the passed positional arguments if
        this is the first instantiation of this class passed these arguments
        *or* simply return the previously instantiated instance of this class
        otherwise (i.e., if this is a subsequent instantiation of this class
        re-passed these same arguments).

        Caveats
        ----------
        This method intentionally accepts *only* positional arguments. See the
        metaclass docstring for further details.

        Parameters
        ----------
        cls : type
            Class whose class is this metaclass.

        All remaining parameters are passed as is to the superclass
        :meth:`type.__call__` method.
        '''

        # Bear witness to the terrifying power of @callable_cached.
        return super().__call__(*args)  # type: ignore[misc]



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype :mod:`typing` **compatibility layer.**

This submodule declares the exact same set of **public typing attributes**
(i.e., module-scoped attributes listed by the :attr:`typing.__all__` global) as
declared by the :mod:`typing` module for your current Python version. Although
the attributes declared by this submodule *mostly* share the same values as
the attributes declared by :mod:`typing`, notable differences include:

* :pep:`585`-deprecated typing attributes. :pep:`585` deprecated **38 public
  typing attributes** to "...be removed from the typing module in the first
  Python version released 5 years after the release of Python 3.9.0." This
  submodule preserves those attributes under their original names for the
  Python 3.8-specific version of the :mod:`typing` module, thus preserving
  forward compatibility with future Python versions. These include:

  * :obj:`typing.AbstractSet`.
  * :obj:`typing.AsyncContextManager`.
  * :obj:`typing.AsyncGenerator`.
  * :obj:`typing.AsyncIterable`.
  * :obj:`typing.AsyncIterator`.
  * :obj:`typing.Awaitable`.
  * :obj:`typing.ByteString`.
  * :obj:`typing.Callable`.
  * :obj:`typing.ChainMap`.
  * :obj:`typing.Collection`.
  * :obj:`typing.Container`.
  * :obj:`typing.ContextManager`.
  * :obj:`typing.Coroutine`.
  * :obj:`typing.Counter`.
  * :obj:`typing.DefaultDict`.
  * :obj:`typing.Deque`.
  * :obj:`typing.Dict`.
  * :obj:`typing.FrozenSet`.
  * :obj:`typing.Generator`.
  * :obj:`typing.ItemsView`.
  * :obj:`typing.Iterable`.
  * :obj:`typing.Iterator`.
  * :obj:`typing.KeysView`.
  * :obj:`typing.List`.
  * :obj:`typing.Mapping`.
  * :obj:`typing.MappingView`.
  * :obj:`typing.Match`.
  * :obj:`typing.MutableMapping`.
  * :obj:`typing.MutableSequence`.
  * :obj:`typing.MutableSet`.
  * :obj:`typing.OrderedDict`.
  * :obj:`typing.Pattern`.
  * :obj:`typing.Reversible`.
  * :obj:`typing.Set`.
  * :obj:`typing.Tuple`.
  * :obj:`typing.Type`.
  * :obj:`typing.Sequence`.
  * :obj:`typing.ValuesView`.

Usage
----------
:mod:`beartype` users are strongly encouraged to import typing attributes from
this submodule rather than from :mod:`typing` directly: e.g.,

.. code-block:: python

   # Instead of this...
   from typing import Tuple, List, Dict, Set, FrozenSet, Type

   # ...always do this.
   from beartype.typing import Tuple, List, Dict, Set, FrozenSet, Type
'''

# ....................{ TODO                               }....................
#FIXME: Fundamentally generalize this submodule to optionally backport
#attributes from "typing_extensions" where available, resolving issue #237 at:
#    https://github.com/beartype/beartype/issues/237
#
#To do so, we'll basically want to discard the entire current implementation of
#this submodule in favour of a fundamentally superior approach resembling:
#    # In "beartype.typing.__init__": the future of typing backports begins today.
#    from typing import TYPE_CHECKING
#
#    # If @beartype is currently being statically type-checked (e.g.,
#    # by mypy or pyright), just defer to the third-party
#    # "typing_extensions" package.
#    #
#    # Note that this does *NOT* mean that @beartype now unconditionally
#    # requires "typing_extensions" at either runtime or static
#    # type-checking time. Any code in an "if TYPE_CHECKING:" is (basically)
#    # just a convincing semantic lie that everything syntactically ignores.
#    if TYPE_CHECKING:
#        from typing_extensions import *  # <-- heh
#    # Else, @beartype is currently being imported from at runtime. This is
#    # the common case. This is also the non-trivial case, because @beartype
#    # does *NOT* require "typing_extensions" as a mandatory runtime
#    # dependency, because @beartype requires *NOTHING* as a runtime
#    # dependency. This is the only rule in @beartype's Rule of Law.
#    else:
#        #FIXME: Unfortunately, to avoid circular import dependencies, these
#        #imports will need to be copy-and-pasted into equivalent condensed
#        #submodules of a new "beartype.typing._util" subpackage.
#        # Import the requisite machinery that will make the magic happen.
#        from beartype._util.hint.utilhintfactory import TypeHintTypeFactory
#        from beartype._util.api.standard.utiltyping import (
#            import_typing_attr_or_fallback as _import_typing_attr_or_fallback)
#
#        # Dynamically define the "Self" type hint as follows:
#        # * If the active Python interpreter targets Python >= 3.11, just
#        #   defer to the canonical "typing.Self" type hint.
#        # * Else if "typing_extensions" is importable *AND* of a sufficiently
#        #   recent version to define the backported "typing_extensions.Self"
#        #   type hint, fallback to that hint.
#        # * Else, synthesize a placeholder type hint that @beartype internally
#        #   recognizes as semantically equivalent to "typing.Self".
#        Self = _import_typing_attr_or_fallback('Self', object)
#        LiteralString = _import_typing_attr_or_fallback('Self', str)
#        TypeGuard = _import_typing_attr_or_fallback('Self', bool)
#        Annotated = _import_typing_attr_or_fallback('Annotated', bool)
#
#        #FIXME: Repeat the above logic for *ALL* existing "typing" attributes.
#FIXME: Actually, ain't nobody got time for that at the moment. The low-hanging
#fruit here is to just trivially alias this submodule to the "typing" module
#like so:
#    from typing import *
#
#Then:
#* Remove almost *ALL* of this submodule.
#* Conditionally redefine various attributes (e.g., deprecated PEP 484 stuff,
#  slow "Protocol" superclass) as needed.
#
#This has the dramatic improvement of innately synchronizing this submodule,
#which has become a *NIGHTMARE* to maintain, against the official
#implementation. Python 3.14 @beartype users hit this when early release
#candidates of Python 3.14 removed "typing.ByteString", only for the final
#release of Python 3.14 to add it back. @beartype then began raising exceptions
#on importation. Since "typing" is a moving target nightmare, exposing ourselves
#to that nightmare no longer makes sense at all.
#
#The only possible issue might be mypy and pyright. If either complain, we'll
#have no choice but to eventually *DEPRECATE* this entire submodule -- probably
#in the run up to @beartype 1.0.0. Ugly stuff, but inviting synchronization woes
#is even uglier.

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To avoid polluting the public module namespace, external attributes
# *NOT* intended for public importation should be locally imported at module
# scope *ONLY* under alternate private names (e.g., "import re as _re" rather
# than merely "from re").
# WARNING: To preserve PEP 561 compliance with static type checkers (e.g.,
# mypy), external attributes *MUST* be explicitly imported with standard static
# import machinery rather than non-standard dynamic import shenanigans (e.g.,
# "from typing import Annotated" rather than
# "import_typing_attr_or_none('Annotated')").
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from beartype._util.py.utilpyversion import (
    IS_PYTHON_AT_MOST_3_16  as _IS_PYTHON_AT_MOST_3_16,
    IS_PYTHON_AT_MOST_3_15  as _IS_PYTHON_AT_MOST_3_15,
    IS_PYTHON_AT_MOST_3_13  as _IS_PYTHON_AT_MOST_3_13,
    IS_PYTHON_AT_LEAST_3_14 as _IS_PYTHON_AT_LEAST_3_14,
    IS_PYTHON_AT_LEAST_3_13 as _IS_PYTHON_AT_LEAST_3_13,
    IS_PYTHON_AT_LEAST_3_12 as _IS_PYTHON_AT_LEAST_3_12,
    IS_PYTHON_AT_LEAST_3_11 as _IS_PYTHON_AT_LEAST_3_11,
)

# ....................{ IMPORTS ~ all                      }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To prevent "mypy --no-implicit-reexport" from raising literally
# hundreds of errors at static analysis time, *ALL* public attributes *MUST* be
# explicitly reimported under the same names with "{exception_name} as
# {exception_name}" syntax rather than merely "{exception_name}". Yes, this is
# ludicrous. Yes, this is mypy. For posterity, these failures resemble:
#     beartype/_cave/_cavefast.py:47: error: Module "beartype.roar" does not
#     explicitly export attribute "BeartypeCallUnavailableTypeException";
#     implicit reexport disabled  [attr-defined]
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# Import all public attributes of the "typing" module both available under all
# supported Python versions and *NOT* deprecated by a subsequent Python version
# under their original names.
from typing import (
    TYPE_CHECKING as TYPE_CHECKING,
    Any as Any,
    Annotated as Annotated,
    BinaryIO as BinaryIO,
    ClassVar as ClassVar,
    Concatenate as Concatenate,  # pyright: ignore
    Final as Final,  # pyright: ignore
    ForwardRef as ForwardRef,
    Generic as Generic,
    IO as IO,
    Literal as Literal,  # pyright: ignore
    NewType as NewType,
    NamedTuple as NamedTuple,
    NoReturn as NoReturn,
    Optional as Optional,
    ParamSpec as ParamSpec,  # pyright: ignore
    ParamSpecArgs as ParamSpecArgs,  # pyright: ignore
    ParamSpecKwargs as ParamSpecKwargs,  # pyright: ignore
    Reversible as Reversible,  # pyright: ignore
    TypedDict as TypedDict,  # pyright: ignore
    Text as Text,
    TextIO as TextIO,
    TypeAlias as TypeAlias,  # pyright: ignore
    TypeGuard as TypeGuard,  # pyright: ignore
    TypeVar as TypeVar,
    Union as Union,
    cast as cast,
    final as final,  # pyright: ignore
    get_args as get_args,  # pyright: ignore
    get_origin as get_origin,  # pyright: ignore
    get_type_hints as get_type_hints,
    is_typeddict as is_typeddict,  # pyright: ignore
    no_type_check as no_type_check,
    no_type_check_decorator as no_type_check_decorator,
    overload as overload,
)

# ....................{ IMPORTS ~ version : at least       }....................
# Import all public attributes of the "typing" module both available under at
# least some Python interpreter version *AN* not yet deprecated by a subsequent
# Python interpreter version under their original names.

#FIXME: mypy is now emitting non-fatal warnings about our failing to import from
#"typing_extensions", which is both an overly strongly opinionated position for
#mypy to stake out *AND* a bad opinion at that, because "typing_extensions" is
#a third-party package. Ideally, mypy shouldn't be pushing *ANY* third-party
#packages. These warnings resemble:
#    beartype/typing/__init__.py:145: note: Use `from typing_extensions import Final` instead
#    beartype/typing/__init__.py:145: note: See https://mypy.readthedocs.io/en/stable/runtime_troubles.html#using-new-additions-to-the-typing-module
#    beartype/typing/__init__.py:145: note: Use `from typing_extensions import Literal` instead
#
#That's not the worst, however. mypy is erroneously ignoring our intentional
#"# type: ignore[attr-defined]" pragmas here. It's likely that the ultimate
#culprit is our use of beartype-specific "IS_PYTHON_AT_LEAST_*" boolean globals.
#Instead, mypy appears to only support hard-coded tests against the
#"sys.version_info" tuple: e.g.,
#    if sys.version_info >= (3, 8):
#
#To resolve this, we should consider:
#* Abandoning our usage of beartype-specific "IS_PYTHON_AT_LEAST_*" boolean
#  globals for hard-coded tests against the "sys.version_info" tuple (above).
#* Submitting an upstream issue requesting that mypy respect the
#  "# type: ignore[attr-defined]" pragma rather than emitting warnings here.

# If the active Python interpreter targets Python >= 3.11...
if _IS_PYTHON_AT_LEAST_3_11:
    from typing import (  # type: ignore[attr-defined]
       LiteralString as LiteralString,  # pyright: ignore
       Never as Never,  # pyright: ignore
       NotRequired as NotRequired,  # pyright: ignore
       Required as Required,  # pyright: ignore
       Self as Self,  # pyright: ignore
       TypeVarTuple as TypeVarTuple,  # pyright: ignore
       Unpack as Unpack,  # pyright: ignore
       assert_never as assert_never,  # pyright: ignore
       assert_type as assert_type,  # pyright: ignore
       clear_overloads as clear_overloads,  # pyright: ignore
       dataclass_transform as dataclass_transform,  # pyright: ignore
       get_overloads as get_overloads,  # pyright: ignore
       reveal_type as reveal_type,  # pyright: ignore
    )

    # If the active Python interpreter targets Python >= 3.12...
    if _IS_PYTHON_AT_LEAST_3_12:
        from typing import (  # type: ignore[attr-defined]
            TypeAliasType as TypeAliasType,  # pyright: ignore
            override as override,  # pyright: ignore
        )

        # If the active Python interpreter targets Python >= 3.13...
        if _IS_PYTHON_AT_LEAST_3_13:
            from typing import (  # type: ignore[attr-defined]
                NoDefault as NoDefault,  # pyright: ignore
                ReadOnly as ReadOnly,  # pyright: ignore
                TypeIs as TypeIs,  # pyright: ignore
                get_protocol_members as get_protocol_members,  # pyright: ignore
                is_protocol as is_protocol,  # pyright: ignore
            )

            # If the active Python interpreter targets Python >= 3.14...
            if _IS_PYTHON_AT_LEAST_3_14:
                from typing import (  # type: ignore[attr-defined]
                    evaluate_forward_ref as evaluate_forward_ref,  # pyright: ignore
                )

# ....................{ IMPORTS ~ version : at most        }....................
# Import all public attributes of the "typing" module both available under at
# most some Python interpreter version (typically due to having been deprecated
# by a prior Python interpreter version).

# If the active Python interpreter targets at most Python <= 3.16...
if _IS_PYTHON_AT_MOST_3_16:
    # Import the PEP 585-compliant "collections.abc.ByteString" attribute
    # under 3.9 <= Python <= 3.13. Both "collections.abc.ByteString" *AND*
    # "typing.ByteString" have been scheduled for removal under Python 3.17 by
    # the upstream CPython issue:
    #     https://github.com/python/cpython/issues/91896
    #
    # Note that these attributes were originally scheduled for removal under
    # Python 3.14. This removal was since deferred by three minor versions (and
    # thus three years) to inform downstream third-party packages with proper
    # "DeprecationWarning" warnings emitted by the "typing" module.
    from collections.abc import ByteString as ByteString  # type: ignore[attr-defined]

    # If the active Python interpreter targets at most Python <= 3.15...
    if _IS_PYTHON_AT_MOST_3_15:
        # Import the PEP 484-compliant "typing.AnyStr" attribute under 3.9 <=
        # Python <= 3.15. This attribute has been scheduled for removal under
        # Python 3.16 by the upstream CPython issue:
        #     https://github.com/python/cpython/issues/105578
        from typing import AnyStr as AnyStr

# ....................{ PEP ~ 544                          }....................
# If this interpreter is performing static type-checking (e.g., via mypy), defer
# to the standard library versions of the family of "Supports*" protocols.
if TYPE_CHECKING:
    from typing import (  # type: ignore[attr-defined]
        Protocol as Protocol,  # pyright: ignore
        SupportsAbs as SupportsAbs,
        SupportsBytes as SupportsBytes,
        SupportsComplex as SupportsComplex,
        SupportsFloat as SupportsFloat,
        SupportsIndex as SupportsIndex,  # pyright: ignore
        SupportsInt as SupportsInt,
        SupportsRound as SupportsRound,
        runtime_checkable as runtime_checkable,  # pyright: ignore
    )
# Else, this interpreter is *NOT* performing static type-checking. In this
# case, prefer our optimized PEP 544 attributes.
else:
    from beartype.typing._typingpep544 import (
        Protocol as Protocol,
        SupportsAbs as SupportsAbs,
        SupportsBytes as SupportsBytes,
        SupportsComplex as SupportsComplex,
        SupportsFloat as SupportsFloat,
        SupportsIndex as SupportsIndex,
        SupportsInt as SupportsInt,
        SupportsRound as SupportsRound,
        runtime_checkable as runtime_checkable,
    )

# ....................{ PEP ~ 585                          }....................
# If the active Python interpreter is performing static type-checking (e.g.,
# "mypy"), import *ALL* public attributes of the "typing" module deprecated by
# PEP 585 as their original values.
#
# This is intentionally performed *BEFORE* the corresponding "else:" branch.
# Why? Because "mypy". If the order of these two branches is reversed, "mypy"
# emits errors when attempting to subscript *ANY* builtin type: e.g.,
#     error: "tuple" is not subscriptable  [misc]
if TYPE_CHECKING:
    from typing import (
        AbstractSet as AbstractSet,
        AsyncContextManager as AsyncContextManager,
        AsyncGenerator as AsyncGenerator,
        AsyncIterable as AsyncIterable,
        AsyncIterator as AsyncIterator,
        Awaitable as Awaitable,
        Callable as Callable,
        ChainMap as ChainMap,
        Collection as Collection,
        Container as Container,
        ContextManager as ContextManager,
        Coroutine as Coroutine,
        Counter as Counter,
        DefaultDict as DefaultDict,
        Deque as Deque,
        Dict as Dict,
        FrozenSet as FrozenSet,
        Generator as Generator,
        Hashable as Hashable,
        ItemsView as ItemsView,
        Iterable as Iterable,
        Iterator as Iterator,
        KeysView as KeysView,
        List as List,
        Mapping as Mapping,
        Match as Match,
        MappingView as MappingView,
        MutableMapping as MutableMapping,
        MutableSequence as MutableSequence,
        MutableSet as MutableSet,
        OrderedDict as OrderedDict,
        Pattern as Pattern,
        Reversible as Reversible,
        Set as Set,
        Sized as Sized,
        Tuple as Tuple,
        Type as Type,
        Sequence as Sequence,
        ValuesView as ValuesView,
    )
# Else, the active Python interpreter is *NOT* performing static type-checking.
# In this case, alias *ALL* public attributes of the "typing" module deprecated
# by PEP 585 to their equivalent values elsewhere in the standard library.
else:
    from collections import (
        ChainMap as ChainMap,
        Counter as Counter,
        OrderedDict as OrderedDict,
        defaultdict as DefaultDict,
        deque as Deque,
    )
    from collections.abc import (
        AsyncIterable as AsyncIterable,
        AsyncIterator as AsyncIterator,
        AsyncGenerator as AsyncGenerator,
        Awaitable as Awaitable,
        Callable as Callable,
        Collection as Collection,
        Container as Container,
        Coroutine as Coroutine,
        Generator as Generator,
        Hashable as Hashable,
        ItemsView as ItemsView,
        Iterable as Iterable,
        Iterator as Iterator,
        KeysView as KeysView,
        Mapping as Mapping,
        MappingView as MappingView,
        MutableMapping as MutableMapping,
        MutableSequence as MutableSequence,
        MutableSet as MutableSet,
        Reversible as Reversible,
        Sequence as Sequence,
        Sized as Sized,
        ValuesView as ValuesView,
        Set as AbstractSet,
    )
    from contextlib import (
        AbstractContextManager as ContextManager,
        AbstractAsyncContextManager as AsyncContextManager,
    )
    from re import (
        Match as Match,
        Pattern as Pattern,
    )

    Dict = dict  # type: ignore[misc]
    FrozenSet = frozenset  # type: ignore[misc]
    List = list  # type: ignore[misc]
    Set = set  # type: ignore[misc]
    Tuple = tuple  # type: ignore[assignment]
    Type = type  # type: ignore[assignment]



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **Python identifier utilities** (i.e., low-level callables handling
unqualified and qualified attribute, callable, class, module, and variable
names).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar._roarexc import _BeartypeUtilTextIdentifierException
from beartype._data.typing.datatyping import TypeException

# ....................{ RAISERS                            }....................
def die_unless_identifier(
    # Mandatory parameters.
    text: str,

    # Optional parameters.
    exception_cls: TypeException = _BeartypeUtilTextIdentifierException,
    exception_prefix: str = '',
) -> None:
    '''
    Raise an exception unless the passed string is a valid **Python attribute
    name** (i.e., ``.``-delimited concatenation of one or more
    :pep:`3131`-compliant syntactically valid Python identifiers, including the
    names of attributes, callables, classes, modules, and variables).

    Parameters
    ----------
    text : str
        String to be validated.
    exception_cls : Type[Exception]
        Type of exception to be raised in the event of a fatal error. Defaults
        to :exc:`._BeartypeUtilTextIdentifierException`.
    exception_prefix : str, optional
        Human-readable label prefixing the representation of this string in the
        exception message. Defaults to the empty string.

    Raises
    ------
    exception_cls
        If this string is *not* a valid Python attribute name.

    See Also
    --------
    :func:`.is_identifier`
        Further details.
    '''

    # If this string is *NOT* a valid Python attribute name, raise an exception.
    if not (isinstance(text, str) and is_identifier(text)):
        assert isinstance(exception_cls, type), (
            'f{repr(exception_cls)} not exception class.')
        assert isinstance(exception_prefix, str), (
            'f{repr(exception_prefix)} not string.')

        raise exception_cls(
            f'{exception_prefix}{repr(text)} not valid Python attribute name.')
    # Else, this string is a valid Python attribute name.

# ....................{ TESTERS                            }....................
def is_dunder(text: str) -> bool:
    '''
    :data:`True` only if the passed string vaguely conforms to the name of a
    **dunder attribute** (i.e., attribute whose name is both prefixed and
    suffixed by ``"__"`` double underscore substrings).

    Parameters
    ----------
    text : str
        String to be inspected.

    Returns
    -------
    bool
        :data:`True` only if this string conforms to a dunder attribute name.
    '''
    assert isinstance(text, str), f'{repr(text)} not string.'

    #FIXME: *INSUFFICIENT.* We also need to call is_identifier() here. *sigh*
    # Return us up the powerful one-liner of power.
    return text.startswith('__') and text.endswith('__')


def is_identifier(text: str) -> bool:
    '''
    :data:`True` only if the passed string is a valid **Python attribute name**
    (i.e., ``.``-delimited concatenation of one or more :pep:`3131`-compliant
    syntactically valid Python identifiers, including the names of attributes,
    callables, classes, modules, and variables).

    This tester is suitable for detecting whether this string is the
    fully-qualified name of an arbitrary Python object.

    Caveats
    -------
    **This tester is mildly slow,** due to unavoidably splitting this string on
    ``.`` delimiters and iteratively passing each of the split substrings to
    the :meth:`str.isidentifier` builtin. Due to the following caveat, this
    inefficiency is unavoidable.

    **This tester is not optimizable with regular expressions** -- at least,
    not trivially. Technically, this tester *can* be optimized by leveraging
    the "General Category" of Unicode filters provided by the third-party
    :mod:`regex` package. Practically, doing so would require the third-party
    :mod:`regex` package and would still almost certainly fail in edge cases.
    Why? Because Python 3 permits Python identifiers to contain Unicode letters
    and digits in the "General Category" of Unicode code points, which is
    extremely non-trivial to match with the standard :mod:`re` module.

    Parameters
    ----------
    text : str
        String to be inspected.

    Returns
    -------
    bool
        :data:`True` only if this string is the ``.``-delimited concatenation of
        one or more syntactically valid Python identifiers.
    '''
    assert isinstance(text, str), f'{repr(text)} not string.'

    # If this text contains *NO* "." delimiters and is thus expected to be an
    # unqualified Python identifier, return true only if this is the case.
    if '.' not in text:
        return text.isidentifier()
    # Else, this text contains one or more "." delimiters and is thus expected
    # to be a qualified Python identifier.

    # Return true only if *ALL* "."-delimited substrings split from this string
    # are valid unqualified Python identifiers. Note that:
    # * Regular expressions report false negatives. See the docstring.
    # * Manual iteration is significantly faster than "all(...)"- and
    #   "any(...)"-style comprehensions.
    # * This approach correctly handles *ALL* edge cases, including when:
    #   * This string is simply the "." character. In this case:
    #         >>> '.'.split('.')
    #         ['', '']
    #     Since the empty string is *NOT* a valid Python identifier, this
    #     iteration immediately returns false as expected.
    # * There exists an alternative and significantly more computationally
    #   expensive means of testing this condition, employed by the
    #   typing.ForwardRef.__init__() method to valid the validity of the passed
    #   relative classname:
    #       # Needless to say, we'll never be doing this.
    #       try:
    #           all(
    #               compile(identifier, '<string>', 'eval')
    #               for identifier in text.split('.')
    #           )
    #           return True
    #       except SyntaxError:
    #           return False
    for text_basename in text.split('.'):
        # If this "."-delimited substring is *NOT* a valid unqualified Python
        # identifier, return false.
        if not text_basename.isidentifier():
            return False
        # Else, this "."-delimited substring is a valid unqualified Python
        # identifier. In this case, silently continue to the next.

    # Return true, since *ALL* "."-delimited substrings split from this string
    # are valid unqualified Python identifiers.
    return True



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **beartype blacklist utilities** (i.e., low-level callables
detecting whether passed objects are blacklisted and thus ignorable with respect
to :mod:`beartype`-specific type-checking, typically due to residing in
third-party packages or modules well-known to be hostile to runtime
type-checking and thus :mod:`beartype`).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype._data.conf.dataconfblack import (
    BLACKLIST_MODULE_NAME_TO_TYPE_NAMES,
    BLACKLIST_PACKAGE_NAMES,
    BLACKLIST_TYPE_MRO_ROOT_MODULE_NAME_TO_TYPE_NAMES,
)
from beartype._util.cache.utilcachecall import callable_cached

# ....................{ TESTERS ~ object                   }....................
@callable_cached
def is_object_blacklisted(obj: object) -> bool:
    '''
    :data:`True` only if the passed object (e.g., callable, class) is
    **beartype-blacklisted** (i.e., resides in a third-party package or modules
    well-known to be hostile to runtime type-checking and thus :mod:`beartype`).

    This tester is memoized for efficiency.

    Parameters
    ----------
    obj : object
        Arbitrary object to be inspected.

    Returns
    -------
    bool
        :data:`True` only if this object is beartype-blacklisted.

    See Also
    --------
    :data:`.BLACKLIST_PACKAGE_NAMES`
        Detailed discussion of beartype-blacklisting.
    '''

    # ....................{ IMPORTS                        }....................
    # Avoid circular import dependencies.
    from beartype._util.module.utilmodget import get_object_module_name_or_none

    # ....................{ PHASES                         }....................
    # This tester is internally implemented as a series of sequential phases --
    # each increasingly more time- and/or space-complex than the last and thus
    # intentionally ordered from least to most complex.

    # ....................{ PHASE ~ type -> module         }....................
    # In this early phase, we efficiently test whether the combination of the
    # fully-qualified name of the module defining the type of the passed object
    # *AND* the unqualified basename of that type is known to be blacklisted.

    # Type of this object.
    obj_type = obj.__class__

    # Fully-qualified name of the package or module defining this object's type
    # if any *OR* "None" otherwise (e.g., if this type is defined in-memory).
    obj_type_module_name = get_object_module_name_or_none(obj_type)

    # If this type defines *NO* module name, this type is *NOT* blacklisted.
    # Why? Because the only types that @beartype blacklists are all defined in
    # modules that physically exist and thus have names. But this type has *NO*
    # module name! In this case, silently reduce to a noop.
    if not obj_type_module_name:
        # print(f'Ignoring unmoduled object {repr(obj)}!')
        return False
    # Else, this type defines this name.

    #FIXME: [SPEED] Globalize the dict.get() bound method called here. *shrug*
    # Frozen set of the unqualified basenames of all beartype-blacklisted types
    # defined by that package or module if any *OR* "None" otherwise (if that
    # package or module defines *NO* beartype-blacklisted types).
    blacklist_obj_type_names = BLACKLIST_MODULE_NAME_TO_TYPE_NAMES.get(
        obj_type_module_name)
    # print(f'obj: {obj}')
    # print(f'obj_type_module_name: {obj_type_module_name}')
    # print(f'blacklist_obj_type_names: {blacklist_obj_type_names}')

    # If...
    if (
        # That package or module defines beartype-blacklisted types *AND*...
        blacklist_obj_type_names and
        # The unqualified basename of this object's type is blacklisted...
        obj_type.__name__ in blacklist_obj_type_names
    ):
        # print(f'Object {obj} blacklisted via "type -> module" heuristic!')

        # Then immediately return true.
        return True
    # Else, this object's type is *NOT* beartype-blacklisted. However, this
    # object could still be beartype-blacklisted in some way. Continue testing!

    # ....................{ PHASE ~ type -> mro -> module  }....................
    # In this early phase, we efficiently test whether the combination of the
    # fully-qualified name of the module defining the type of the passed object
    # *AND* the unqualified basename of that type is known to be blacklisted
    # such that that type masquerades as the low-level user-defined callable it
    # wraps and is thus *ONLY* accessible as the root method-resolution order
    # (MRO) item (i.e., second-to-last item of the "__mro__" dunder dictionary
    # of this type, thus ignoring the ignorable "object" guaranteed to be the
    # last item of all such dictionaries).
    #
    # @beartype doesn't make the rules. It only complains about and breaks them.

    # MRO of this type.
    #
    # Note that all types are guaranteed to have a root MRO item *EXCEPT* the
    # "object" superclass, whose simplistic "(object,)" MRO lacks a root item.
    # While we could manually exclude this superclass, the existence of even a
    # single exception to this guarantee suggests that devious users could
    # circumvent this guarantee... somehow. Users are devious. Who can fathom
    # their ways? For safety, we assume this guarantee to *NOT* globally hold.
    obj_type_mro = obj_type.__mro__

    # If this MRO contains two or more items, this type is *NOT* the trivial
    # "object" superclass or something like that superclass. In this case...
    if len(obj_type_mro) >= 2:
        # Root MRO item of this type, ignoring the trivial "object" superclass.
        obj_type_mro_root = obj_type_mro[-2]

        # Fully-qualified name of the package or module defining this object's
        # root MRO type if any *OR* "None" otherwise (e.g., if this type is
        # defined in-memory).
        obj_type_mro_root_module_name = get_object_module_name_or_none(
            obj_type_mro_root)

        # If this type defines this name...
        if obj_type_mro_root_module_name:
            #FIXME: [SPEED] Globalize the dict.get() bound method called here.
            # Frozen set of the unqualified basenames of all
            # beartype-blacklisted types defined by that package or module if
            # any *OR* "None" otherwise (if that package or module defines *NO*
            # beartype-blacklisted types).
            blacklist_obj_type_mro_root_type_names = (
                BLACKLIST_TYPE_MRO_ROOT_MODULE_NAME_TO_TYPE_NAMES.get(
                    obj_type_mro_root_module_name))
            # print(f'obj: {obj}')
            # print(f'obj_type_mro_root_module_name: {obj_type_mro_root_module_name}')
            # print(f'blacklist_obj_type_mro_root_type_names: {blacklist_obj_type_mro_root_type_names}')

            # If...
            if (
                # That package or module defines beartype-blacklisted types
                # *AND*...
                blacklist_obj_type_mro_root_type_names and
                # The unqualified basename of this object's type is
                # blacklisted...
                obj_type_mro_root.__name__ in (
                    blacklist_obj_type_mro_root_type_names)
            ):
                # print(f'Object {obj} blacklisted via "type -> mro -> module" heuristic!')

                # Then immediately return true.
                return True
            # Else, this object's type is *NOT* beartype-blacklisted. However,
            # this object could still be beartype-blacklisted in some way.
            # Continue testing!
        # Else, this type defines *NO* module name. This object's type is *NOT*
        # beartype-blacklisted. However, this object could still be
        # beartype-blacklisted in some way. Continue testing!
    # Else, this type is the trivial "object" superclass or something like that
    # superclass.

    # ....................{ PHASE ~ package                }....................
    # In this late phase, we inefficiently test whether the combination of the
    # fully-qualified name of the top-level root package directly defining the
    # passed object is known to be blacklisted. This heuristic is less
    # efficient, as stripping this package name from this module name
    # constitutes a string-munging operation.

    # Fully-qualified name of the package or module defining this object if any
    # *OR* "None" otherwise (e.g., if this object is defined in-memory).
    obj_module_name = get_object_module_name_or_none(obj)

    # If this object defines *NO* module name, silently reduce to a noop.
    if not obj_module_name:
        # print(f'Ignoring unmoduled object {repr(obj)}!')
        return False
    # Else, this object defines this name and is thus *PROBABLY* either a
    # pure-Python class or callable.

    # Fully-qualified name of the top-level root package or module transitively
    # containing that package or module (e.g., "some_package" when
    # "obj_module_name" is "some_package.some_module.some_submodule").
    #
    # Note this has been profiled to be the fastest one-liner for parsing the
    # first "."-suffixed substring from a "."-delimited string.
    obj_package_name = obj_module_name.partition('.')[0]
    # print(f'Testing package {repr(obj_package_name)} for blacklisting...')

    # If this package is globally beartype-blacklisted, immediately return true.
    if obj_package_name in BLACKLIST_PACKAGE_NAMES:
        return True
    # Else, this package is *NOT* globally beartype-blacklisted. However, this
    # object could still be specifically beartype-blacklisted. Continue testing!

    # Return false as a feeble fallback.
    return False



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype Inferential Type-hint Engine (BITE) type hint inferrers** (i.e.,
high-level functions dynamically inferring the type hints best describing
arbitrary objects).
'''

# ....................{ IMPORTS                            }....................
from beartype.bite.kind.infercallable import infer_hint_callable
from beartype.bite.kind.inferthirdparty import infer_hint_thirdparty
from beartype.bite.collection.infercollectionbuiltin import (
    infer_hint_collection_builtin)
from beartype.bite.collection.infercollectionsabc import (
    infer_hint_collections_abc)
from beartype.roar import BeartypeDoorInferHintRecursionWarning
from beartype.typing import (
    Callable,
    Optional,
    Tuple,
    Type,
)
from beartype._conf.confmain import BeartypeConf
from beartype._conf.confcommon import get_beartype_conf_strategy_on
from beartype._conf.conftest import die_unless_conf
from beartype._data.cls.datacls import TYPES_BUILTIN_SCALAR
from beartype._data.typing.datatyping import FrozenSetInts
from beartype._data.kind.datakindset import FROZENSET_EMPTY
from beartype._util.error.utilerrwarn import issue_warning
from beartype._util.hint.pep.utilpeptest import is_hint_pep
from beartype._util.text.utiltextrepr import represent_object

# ....................{ CLASSES                            }....................
class BeartypeInferHintContainerRecursion(object):
    '''
    Child type hint subscripting all **recursive container type hints** (i.e.,
    parent type hints describing a container containing one or more items
    self-referentially referring to the same container).
    '''

    def __repr__(self) -> str:
        return 'RecursiveContainer'

# ....................{ INFERERS                           }....................
def infer_hint(
    # Mandatory parameters.
    obj: object,

    # Optional keyword-only parameters.
    *,
    conf: Optional[BeartypeConf] = None,

    # Optional keyword-only hidden parameters. *GULP*
    __beartype_obj_ids_seen__: FrozenSetInts = FROZENSET_EMPTY,
) -> object:
    '''
    Type hint annotating the passed object.

    This function dynamically infers (i.e., computes, decides, deduces) a type
    hint sufficient for annotating (i.e., describing, matching, validating) the
    passed object.

    Caveats
    -------
    **This function explicitly guards against infinite recursion.** Notably,
    this function accepts **recursive containers** (i.e., pure-Python containers
    containing one or more items whose values self-referentially refer to the
    same containers). When passed a recursive container, this function guards
    against infinite recursion that would otherwise be induced by that container
    by instead returning a placeholder instance of the
    :class:`.BeartypeInferHintRecursion` class describing this recursion: e.g.,

    .. code-block:: python

       # Define a trivial recursive list.
       >>> recursive_list = ['this is fine', b'this is fine too, but...',]
       >>> recursive_list.append(recursive_list)

       # Infer the type hint annotating this list.
       >>> from beartype.bite import infer_hint
       >>> infer_hint(recursive_list)
       ##FIXME: INSERT SANE REPR HERE, PLEASE. *sigh*

    **This function exhibits best-, average-, and worst-case** :math:`O(n)`
    **linear-time complexity** for :math:`n` the number of nested items
    transitively contained in the passed object if this object is a container.
    This differs from most of the public :mod:`beartype` API, which exhibits at
    worst worst-case amortized :math:`O(1)` constant-time complexity. Although
    this function could certainly be generalized to support that sort of
    constant-time complexity, doing so would be largely pointless. Type hint
    inference only has real-world value insofar as it accurately infers hints.
    Inference operating in :math:`O(1)` time would necessarily be inaccurate and
    thus of *no* real-world value. Inference operating in :math:`O(n)` time, on
    the other hand, accurately infers hints for even arbitrarily large
    pure-Python containers by introspecting all objects transitively reachable
    from those containers.

    **This function cannot be reasonably memoized** (e.g., via the
    :func:`beartype._util.cache.utilcachecall.callable_cached` decorator).
    Technically, this function *could* be memoized... *sorta.* Pragmatically,
    this function *cannot* be memoized. Most arbitrary objects are mutable and
    thus unhashable and thus unmemoizable. While feasible, memoizing this
    function for the small subset of arbitrary objects that are immutable would
    dramatically increase the space complexity of this function. In short,
    memoizing arbitrary objects is effectively infeasible.

    Parameters
    ----------
    obj : object
        Arbitrary object to infer a type hint from.
    conf : BeartypeConf, optional
        **Beartype configuration** (i.e., self-caching dataclass encapsulating
        all settings configuring type-checking for the passed object). Defaults
        to ``BeartypeConf(strategy=BeartypeStrategy.On))``, the default
        :math:`O(n)` linear-time configuration. Why not the default
        :math:`O(1)` constant-time configuration like the remainder of the
        :mod:`beartype` codebase? Because type hints are *typically* only useful
        when perfectly describing the internal structure of objects; type hints
        that imperfectly describe the internal structure of objects induce
        **false positives** (i.e., cause both static and runtime type-checkers
        to improperly emit errors and raise exceptions for otherwise valid
        objects that would have satisfied those hints had those hints more
        perfectly described the internal structure of those objects). Exceptions
        do exist, however. Downstream third-party consumers that only call this
        function to create an temporary in-memory type hint that is then passed
        to other runtime type-checking functionality (e.g.,
        :func:`beartype.bite.is_bearable`, :func:`beartype.bite.is_subhint`)
        often benefits from :math:`O(1)` constant-time type hint inference.
    __beartype_obj_ids_seen__ : FrozenSet[int]
        **Recursion guard** (i.e., frozen set of the integers uniquely
        identifying all previously visited containers passed as the ``obj``
        parameter to some recursive parent call of this same function on the
        current call stack). If the integer identifying a passed object already
        resides in this recursion guard, that object has already been visited by
        a prior call to this function in the same call stack and is thus a
        recursive container; in that case, this function short-circuits infinite
        recursion by returning a placeholder instance of the
        :class:`.BeartypeInferHintRecursion` class describing this issue.

    Returns
    -------
    object
        Type hint inferred from the passed object.

    Warns
    -----
    BeartypeDoorInferHintRecursionWarning
        On detecting that the passed object is **recursive** (i.e.,
        self-referentially refers to itself, typically due to being a container
        containing one or more items that self-referentially refer to that same
        container).

    Raises
    ------
    BeartypeConfException
        If the passed ``conf`` parameter is *not* a beartype configuration.
    '''

    # ....................{ PREAMBLE                       }....................
    # If the integer uniquely identifying this object already resides in this
    # recursion guard, this object has already been visited by a prior call to
    # this function in the same call stack and is thus a recursive container.
    # In this case...
    if id(obj) in __beartype_obj_ids_seen__:
        # Emit a non-fatal warning informing the caller.
        issue_warning(
            cls=BeartypeDoorInferHintRecursionWarning,
            message=(
                f'Container recursion detected; short-circuiting for safety. '
                f'Container {represent_object(obj)} self-referentially '
                f'contains itself as an item.'
            ),
        )

        # Short-circuit infinite recursion by creating and returning a
        # placeholder instance of a dataclass describing this situation.
        return BeartypeInferHintContainerRecursion
    # Else, this object has yet to be visited.

    # If the caller explicitly passed *NO* configuration, default to the default
    # linear-time configuration.
    if conf is None:
        conf = get_beartype_conf_strategy_on()
    # Else, the caller explicitly passed a configuration.

    # If this configuration is invalid, raise an exception.
    die_unless_conf(conf)
    # Else, this configuration is valid.

    # ....................{ PEP                            }....................
    #FIXME: Generalize to support iterable parametrized generics (e.g.,
    #instances of "Generic[T]" subclasses); iterable parametrized generics
    #should be subscripted by a child type hint describing the items contained
    #by those iterables. For example, when this function is passed an instance
    #of a "List[T]" subclass, this function should return "List[hints_child]"
    #where "hints_child" is the union of the types of all items of this list.
    #
    #Doing so will require explicitly detecting generics here. *sigh*

    # If this object is a PEP-compliant type hint, this object is trivially
    # satisfied by itself. Interestingly:
    # * This includes the "None" singleton, which is a PEP 484-compliant type
    #   hint whose beartype sign is "HintSignNone".
    #
    # Note that PEP-compliant type hints are intentionally detected and
    # short-circuited first *BEFORE* any further inference. Why? Because
    # subjecting PEP-compliant type hints to any inference typically destroys
    # those hints. For example:
    # * Many PEP-compliant type hints are types. Subjecting these hints to
    #   further inference would incorrectly infer the hints for these hints as
    #   "Type[obj]" rather than "obj".
    #
    # Specifically, if...
    if (
        # This object is a PEP-compliant type hint *AND*...
        is_hint_pep(obj) and
        # This object is *NOT* a string. There exists an ambiguity here. Under
        # PEP 484, any string that is a type hint is a stringified forward
        # reference to another PEP-compliant hint that typically has yet to be
        # defined. However, unconditionally inferring *ALL* strings to be PEP
        # 484-compliant stringified forward references here would coerce
        # container items that are simple strings into references. Since doing
        # so would strongly conflict with common sense and sane semantics, this
        # function preserves strings as simple PEP-noncompliant objects.
        not isinstance(obj, str)
    ):
        # Return this PEP-compliant type hint as is.
        return obj
    # Else, this object is *NOT* a PEP-compliant type hint.

    # ....................{ PEP [484|585]                  }....................
    # If this object is a type, this type is trivially satisfied by a PEP 484-
    # or 585-compliant subclass type hint subscripted by this type.
    elif isinstance(obj, type):
        return Type[obj]
    # Else, this object is *NOT* a type.
    #
    # If this object is callable, defer to this lower-level function inferring a
    # "typing.Callable[...]" type hint from this callable.
    elif callable(obj):
        return infer_hint_callable(obj)
    # Else, this object is uncallable.

    # ....................{ NON-PEP ~ scalar               }....................
    # Type of this object.
    obj_type = obj.__class__

    # If this object is a builtin scalar (e.g., integer, string), return this
    # type as is.
    #
    # Note that this is *NOT* simply a negligible optimization, although it
    # certainly is that as well. This is an essential sanity check to ensure
    # that strings are annotated as the builtin "str" type rather than the
    # recursive "collections.abc.Sequence[str]" type hint, which they otherwise
    # would be. Since strings are infinitely recursively immutable sequences of
    # immutable sequences, this detection avoids *INSANE* infinite recursion.
    if obj_type in TYPES_BUILTIN_SCALAR:
        return obj_type
    # Else, this object is *NOT* a builtin scalar.

    # ....................{ INFER                          }....................
    # For each lower-level hint inferer...
    for hint_inferer in _HINT_INFERERS:
        # print(f'Inferring {repr(obj)} hint via inferer {repr(hint_inferer)}...')

        # Hint inferred by this inferer as validating this object if this
        # inferer inferred such a hint *OR* "None" otherwise (i.e., if this
        # inferer failed to infer a hint for this object).
        hint = hint_inferer(  # type: ignore[call-arg]
            obj=obj,
            conf=conf,
            __beartype_obj_ids_seen__=__beartype_obj_ids_seen__,
        )

        # If this inferer inferred a hint for this object, return this hint.
        if hint is not None:
            return hint
        # Else, this inferer inferred *NO* hint for this object. In this case,
        # silently continue to the next hint inferer and hope for better joy.
    # Else, *NO* inferer inferred a hint for this object. We tried, people.

    # ....................{ FALLBACK                       }....................
    # Return this type as a last-ditch fallback. By definition, *ANY* object is
    # trivially satisfied by a type hint that is the type of that object (e.g.,
    # the integer "42" is trivially satisfied by the builtin type "int").
    return obj_type

# ....................{ PRIVATE ~ constants                }....................
_HINT_INFERERS: Tuple[Callable, ...] = (
    # Builtin collections should typically (but *NOT* always, interestingly) be
    # annotated as builtin collection type hints. For example:
    # * Lists satisfy the "collections.abc.MutableSequence" protocol, but are
    #   better annotated as fine-grained "list[...]" type hints rather than as
    #   coarse-grained "collections.abc.MutableSequence[...]" type hints.
    #
    # Exceptions include:
    # * Dictionary keys views (e.g., "{'ugh': 42}.keys()"), which are actually
    #   set instances but still better annotated as fine-grained
    #   "collections.abc.KeysView[...]" type hints rather than as coarse-grained
    #   "set[...]" type hints.
    #
    # This inferer is intentionally listed first to force this prioritization.
    infer_hint_collection_builtin,

    # Non-standard popular third-party objects (e.g., NumPy arrays, PyTorch
    # tensors) should often (but *NOT* always, interestingly) be annotated as
    # PEP 593-compliant "typing(|_extensions).Annotated[...]" type hints
    # inferring both the types *AND* well-known metadata associated with those
    # types (e.g., dimensionality, dtypes, shapes).
    #
    # This inferer is intentionally listed next to force this prioritization.
    infer_hint_thirdparty,

    # User-defined collections satisfying standard "collections.abc" protocols
    # should often (but *NOT* always, interestingly) be annotated as
    # "collections.abc" type hints.
    #
    # This inferer is intentionally listed last to deprioritize this form of
    # coarse-grained inference in favour of more fine-grained inferers.
    infer_hint_collections_abc,
)
'''
Tuple of all **type hint inferers** (i.e., lower-level functions inferring the
type hint for the passed object according to some function-specific heuristic).

Each item of this set is a function accepting an arbitrary object ``obj`` and
the recursion guard ``__beartype_obj_ids_seen__`` and returning either the type
hint inferred validating that object *or* :data:`None` if that function failed
to infer such a type hint. Specifically, each function has a signature
resembling:

.. code-block:: python

   def infer_hint_{heuristic}(
       obj: object, __beartype_obj_ids_seen__: FrozenSetInts) -> Optional[object]:
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`646`-compliant **type hint reducers** (i.e., low-level
callables converting :pep:`646`-compliant hints to lower-level type hints more
readily consumable by :mod:`beartype`).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ TODO                               }....................
#FIXME: Currently, we only shallowly type-check PEP 646-compliant mixed
#fixed-variadic tuple hints as... tuples. It's not much. Obviously, we need to
#deeply type-check these tuple data structures as soon as feasible. There exist
#two distinct use cases here:
#* Fixed-variadic tuple hints containing exactly one unpacked child tuple hint
#  (e.g., "tuple[int, *tuple[str, ...], float]"). Note that a tuple hint may
#  only contain *AT MOST* one unpacked child tuple hint.
#* Fixed-variadic tuple hints containing exactly one type variable tuple, either
#  as:
#  * The first child hint (e.g., "tuple[*Ts, int]").
#  * The last child hint (e.g., "tuple[str, bytes, *Ts]"). Note that this case
#    has a prominent edge case. Fixed-variadic tuple hints of the form
#    "tuple[hint_child, *Ts]" for *ANY* "hint_child" and type variable tuple
#    "Ts" trivially reduce to variadic tuple hints of the form
#    "tuple[hint_child, ...]" at the moment, as we silently ignore *ALL* type
#    variable tuples. Ergo, if a hint is a fixed-variadic tuple hint whose last
#    child hint is a type variable tuple, this hint *MUST* by definition be
#    prefixed by two or more child hints that are *NOT* type variable tuples.
#  * Any child hint other than the first or last (e.g.,
#    "tuple[float, *Ts, bool]").
#
#  Note that:
#  * A parent tuple hint can contain at most *ONE* unpacked child tuple hint.
#    So, we'll now need to record the number of unpacked child tuple hints that
#    have been previously visited and raise an exception if two or more are
#    seen. Ugh!
#  * Again, these cases have a prominent edge case. Fixed-variadic tuple hints
#    of the form "tuple[*Ts]" for *ANY* type variable tuple "Ts" trivially
#    reduce to the builtin type "tuple" at the moment, as we silently ignore
#    *ALL* type variable tuples.
#  * Fixed-variadic tuple hints *CANNOT* contain both an unpacked type variable
#    tuple *AND* unpacked child tuple hint (e.g., "tuple[*Ts, *tuple[int,
#    ...]]"). Fixed-variadic tuple hints can contain at most one unpacked child
#    hint. Hopefully, this constraint reduces the complexity of code generation.
#
#The first place to start with all of this is implementing a new code generation
#algorithm for the new "HintSignPep646TupleFixedVariadic" sign, which currently
#just shallowly reduces to the builtin "tuple" type. Obviously, that's awful.
#The first-draft implementation of this algorithm should just focus on
#fixed-variadic tuple hints containing exactly one type variable tuple (e.g.,
#"tuple[float, *Ts, bool]") for now, as that's the simpler use case. Of course,
#even that's *NOT* simple -- but it's a more reasonable start than unpacked
#child tuple hints, which spiral into madness far faster and harder.
#
#Lastly, note that we can trivially handle unpacked child tuple hints in a
#simple, effective way *WITHOUT* actually investing any effort in doing so. How?
#By simply treating each unpacked child tuple hint as a type variable tuple
#(e.g., by treating "tuple[str, *tuple[int, ...], bytes]" as equivalent to
#"tuple[str, *Ts, bytes]"). Since we already need to initially handle type
#variable tuples anyway, we shatter two birds with one hand. Yes! Yes!
#FIXME: Actually, it's *DEFINITELY* not the case that we "silently ignore *ALL*
#type variable tuples." We reduce both type variables and type variable tuples
#to lookup tables. So, stop ignoring type variable tuples below, please.
#
#Otherwise, everything above seems great -- by which I mean, exhausting.

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeDecorHintPep646Exception
from beartype.typing import Optional
from beartype._data.typing.datatypingport import (
    Hint,
    ListHints,
    TupleHints,
)
from beartype._data.hint.sign.datahintsigns import (
    HintSignPep646TupleUnpacked,
    HintSignPep646TypeVarTupleUnpacked,
)
from beartype._data.hint.sign.datahintsignset import (
    HINT_SIGNS_PEP646_TUPLE_HINT_CHILD_UNPACKED)
from beartype._util.hint.pep.proposal.pep484585646 import (
    is_hint_pep484585646_tuple_variadic,
    make_hint_pep484585_tuple_fixed,
)
from beartype._util.hint.pep.utilpepget import get_hint_pep_args
from beartype._util.hint.pep.utilpepsign import get_hint_pep_sign_or_none

# ....................{ REDUCERS                           }....................
#FIXME: Unit test us up, please.
def reduce_hint_pep646_tuple(
    hint: Hint, exception_prefix: str, **kwargs) -> Hint:
    '''
    Reduce the passed :pep:`646`-compliant **tuple hint** (i.e., parent tuple
    hints subscripted by either a :pep:`646`-compliant type variable tuples *or*
    :pep:`646`-compliant unpacked child tuple hint) to a lower-level type hint
    currently supported by :mod:`beartype`.

    This reducer is intentionally *not* memoized (e.g., by the
    ``callable_cached`` decorator), as reducers cannot be memoized.

    Parameters
    ----------
    hint : object
        :pep:`646`-compliant tuple hint to be reduced.
    exception_prefix : str
        Human-readable substring prefixing raised exception messages.

    All remaining keyword-only parameters are silently ignored.

    Returns
    -------
    Hint
        Lower-level hint currently supported by :mod:`beartype`.

    Raises
    ------
    BeartypeDecorHintPep646Exception
        If this tuple hint is subscripted by either:

        * Two or more :pep:`646`-compliant type variable tuples.
        * Two or more :pep:`646`-compliant unpacked child tuple hints.
        * Two or more :pep:`646`-compliant type variable tuples.
        * A :pep:`646`-compliant type variable tuple *and* an unpacked child
          tuple hint.
    '''
    # print(f'Reducing PEP 646 tuple hint {repr(hint)}...')

    # ....................{ LOCALS                         }....................
    # Tuple of the one or more child hints subscripting this parent tuple hint.
    #
    # Note that the previously called
    # get_hint_pep484585646_tuple_sign_unambiguous() getter responsible for
    # disambiguating PEP 484- and 585-compliant tuple hints from this PEP
    # 646-compliant tuple hint has already pre-validated this tuple hint to be
    # subscripted by two or more child hints.
    hint_childs = get_hint_pep_args(hint)

    # Number of child hints subscripting this parent tuple hint.
    hint_childs_len = len(hint_childs)

    # Assert this parent tuple hint is subscripted by at least one child hint.
    #
    # Note that the previously called
    # get_hint_pep484585646_tuple_sign_unambiguous() getter should have already
    # guaranteed this. Ergo, we avoid raising a full-blown exception here.
    assert hint_childs_len >= 1, (
        f'PEP 646 tuple hint {repr(hint)} subscripted by no child hints.')

    # List of the zero or more new child hints with which to subscript a new
    # PEP 585-compliant parent tuple hint to reduce this PEP 646-compliant
    # parent tuple hint to if this PEP 646-compliant parent tuple hint is
    # reducible to a PEP 585-compliant parent tuple hint *OR* "None" otherwise.
    #
    # This list enables this reducer to reduce PEP 646-compliant parent tuple
    # hints to semantically equivalent PEP 585-compliant parent tuple hints.
    # Generating type-checking code and violation messages for PEP 585-compliant
    # parent tuple hints is both faster *AND* simpler than generating
    # type-checking code and violation messages for PEP 646-compliant parent
    # tuple hints, incentivizing this reduction.
    hint_pep585_childs: Optional[TupleHints] = None

    # ....................{ LENGTH                         }....................
    # If this parent tuple hint is subscripted by exactly one child hint...
    if hint_childs_len == 1:
        # Child hint subscripting this parent tuple hint.
        hint_child = hint_childs[0]

        # Sign uniquely identifying the child hint subscripting this parent
        # tuple hint if this child hint is PEP-compliant *OR* "None" otherwise.
        hint_child_sign = get_hint_pep_sign_or_none(hint_child)
        # print(f'hint_child_sign: {hint_child_sign}')

        # If this child hint is a PEP 646-compliant unpacked type variable tuple
        # (e.g., the "*Ts" in "tuple[*Ts]"), reduce this PEP 646-compliant tuple
        # hint to the builtin "tuple" type.
        #
        # The justification here is somewhat subtle. @beartype currently ignores
        # all type parameters -- including both PEP 484-compliant type variables
        # and 646-compliant type variable tuples. Whereas a conventional type
        # variable is trivially reducible to the ignorable "typing.Any"
        # singleton, however, type variable tuples are only reducible to the
        # less ignorable PEP 646-compliant unpacked child tuple hint
        # "*tuple[typing.Any, ...]". Type variable tuples imply *VARIADICITY*
        # (i.e., a requirement that zero or more tuple items be matched). This
        # requirement *CANNOT* be trivially ignored. Ergo, any PEP 646-complaint
        # parent tuple hint of the form "tuple[*Ts]" for *ANY* type variable
        # tuple "*Ts" is reducible to the also PEP 646-compliant parent tuple
        # hint "tuple[*tuple[typing.Any, ...]]", which unpacks to the PEP
        # 646-*AGNOSTIC* parent tuple hint "tuple[typing.Any, ...]", which then
        # simply reduces to the builtin "tuple" type.
        if hint_child_sign is HintSignPep646TypeVarTupleUnpacked:
            # Reduce this PEP 646-compliant tuple hint to the "tuple" type.
            return tuple
        # Else, this child hint is *NOT* a PEP 646-compliant unpacked type
        # variable tuple.
        #
        # If this child hint is a PEP 646-compliant unpacked child tuple hint
        # (e.g., the "*tuple[str, ...]" in "tuple[*tuple[str, ...]]"), reduce
        # this PEP 646-compliant tuple hint to the semantically equivalent PEP
        # 585-compliant tuple hint subscripted by the child child hints
        # subscripting this unpacked child tuple hint (e.g., from
        # "tuple[*tuple[str, ...]]" to "tuple[str, ...]"). This could be
        # regarded as a non-recursive "unboxing" or "unpacking" operation.
        #
        # Note that this edge case is syntactically permitted by PEP 646 for
        # orthogonality, despite being semantically superfluous and thus
        # conveying *NO* meaningful typing not already conveyed by the simpler
        # PEP 585-compliant tuple hint that this PEP 646-compliant tuple hint
        # reduces to. So it goes, Pythonistas. So it goes.
        elif hint_child_sign is HintSignPep646TupleUnpacked:
            # Reduce this PEP 646-compliant tuple hint to the semantically
            # equivalent PEP 585-compliant tuple hint subscripted by the zero or
            # more child child hints subscripting this unpacked child tuple
            # hint.
            #
            # Note that the CPython parser itself prevents unpacked child tuple
            # hints from being trivially subscripted by *NO* child child hints.
            # Interestingly, doing so is still non-trivially feasible by the
            # standard empty tuple "()" trick: e.g.,
            #     >>> tuple[*tuple[]]
            #                ^^^^^^^
            #     SyntaxError: invalid syntax. Perhaps you forgot a comma?
            #
            #     >>> tuple[*tuple[()]]
            #     tuple[*tuple[()]]
            hint_pep585_childs = get_hint_pep_args(hint_child)
        # Else, this child hint is *NOT* a PEP 646-compliant unpacked child
        # tuple hint. In this case, raise a fatal exception. Why? Because the
        # previously called get_hint_pep484585646_tuple_sign_unambiguous()
        # getter already validated this tuple hint to be PEP 646-compliant and
        # thus be subscripted by one or more PEP 646-compliant child hints, but
        # this hint is subscripted by only one PEP 646-noncompliant child hint!
        else:  # pragma: no cover
            raise BeartypeDecorHintPep646Exception(
                f'{exception_prefix}PEP 646 tuple type hint {repr(hint)} '
                f'child hint {repr(hint_child)} not PEP 646-compliant '
                f'(i.e., neither unpacked type variable tuple nor '
                f'unpacked child tuple type hint).'
            )
    # Else, this parent tuple hint is *NOT* subscripted by one child hint,
    # implying this parent tuple hint is subscripted by two or more child hints.
    # Since this hint is irreducible, preserve this hint as is.

    # ....................{ SEARCH                         }....................
    # If prior logic has *NOT* already decided this PEP 646-compliant parent
    # tuple hint to be trivially reducible to a PEP 585-compliant parent tuple
    # hint, attempt to non-trivially decide this with iteration over all child
    # hints subscripting this parent tuple hint.
    if hint_pep585_childs is None:
        # First PEP 646-compliant child hint subscripting this parent tuple hint
        # discovered by iteration below if this parent tuple hint is subscripted
        # by one or more PEP 646-compliant child hints *OR* "None" otherwise
        # (i.e., if this parent tuple hint is subscripted by *NO* such hints).
        #
        # Note that PEP 646-compliant child hints include:
        # * PEP 646-compliant unpacked type variable tuple (e.g., the "*Ts" in
        #   "tuple[str, *Ts]").
        # * PEP 646-compliant unpacked type variable tuple (e.g., the "*Ts" in
        #   "tuple[str, *Ts]").
        #
        # This local enables the iteration below to validate that this parent
        # tuple hint fully complies with PEP 646 by being subscripted by:
        # * Exactly one PEP 646-compliant child hint (e.g., "*Ts",
        #   "*tuple[int]").
        # * Zero or more PEP 646-noncompliant child hints (e.g., "int",
        #   "set[str]").
        #
        # PEP 646 prohibits parent tuple hints from being subscripted by two or
        # more PEP 646-compliant child hints.
        hint_child_pep646: Optional[Hint] = None

        # 0-based index of this first PEP 646-compliant child hint in this
        # parent tuple hint if any *OR* "None" otherwise.
        hint_child_pep646_index: Optional[int] = None

        #FIXME: [SPEED] Acquire and release a cached list instead, please.
        # Either:
        # * If this hint is subscripted somewhere by a PEP 646-compliant
        #   unpacked child fixed-length tuple hint (e.g., the "*tuple[str,
        #   float]" in "tuple[int, *tuple[str, float]]") and thus semantically
        #   reducible to a PEP 585-compliant parent fixed-length tuple hint
        #   (e.g., from "tuple[int, *tuple[str, float]]" to "tuple[int, str,
        #   float]"), the list of zero or more child hints with which to
        #   subscript a PEP 585-compliant parent fixed-length tuple hint to
        #   reduce this PEP 646-compliant parent tuple hint to.
        # * Else (i.e., if this PEP 646-compliant hint is *NOT* reducible to
        #   such a PEP 585-compliant hint), "None".
        #
        # PEP 646-compliant unpacked child fixed-length tuple hints are entirely
        # superfluous and supported by PEP 646 merely for orthogonality.
        # Although it is unlikely that *ANYONE* will everyone use PEP
        # 646-compliant unpacked child fixed-length tuple hints, the possibility
        # that someone might requires us to support this obscure edge case.
        hint_pep585_childs_list: Optional[ListHints] = []

        #FIXME: [SPEED] Optimize into a "while" loop, please. *sigh*
        # For the 0-based index of each child hint subscripting this parent
        # tuple hint as well as that child hint...
        for hint_child_index, hint_child in enumerate(hint_childs):
            # Sign uniquely identifying this child hint if this child hint is
            # PEP-compliant *OR* "None" otherwise.
            hint_child_sign = get_hint_pep_sign_or_none(hint_child)

            # If this child hint is PEP 646-compliant...
            if hint_child_sign in HINT_SIGNS_PEP646_TUPLE_HINT_CHILD_UNPACKED:
                # If this iteration has yet to discover a PEP 646-compliant
                # child hint of this parent tuple hint, this is the first PEP
                # 646-compliant child hint subscripting this parent tuple hint
                # discovered by this iteration. In this case, record this fact.
                if hint_child_pep646 is None:
                    hint_child_pep646 = hint_child
                    hint_child_pep646_index = hint_child_index

                    # If...
                    if (
                        # This list still exists, it is still unknown whether
                        # this hint is subscripted anywhere by a PEP
                        # 646-compliant unpacked child hint -- implying that
                        # this hint *COULD* still be subscripted somewhere by a
                        # PEP 646-compliant unpacked child fixed-length tuple
                        # hint (e.g., the "*tuple[str, float]" in "tuple[int,
                        # *tuple[str, float]]") *AND*...
                        hint_pep585_childs_list is not None and
                        # This child hint is a PEP 646-compliant unpacked child
                        # tuple hint *AND*...
                        hint_child_sign is HintSignPep646TupleUnpacked and
                        # This child hint is *NOT* a PEP 646-compliant unpacked
                        # child variable-length tuple hint, this child hint
                        # *MUST* by elimination be a PEP 646-compliant unpacked
                        # child fixed-length tuple hint.
                        not is_hint_pep484585646_tuple_variadic(hint_child)
                    ):
                        # Tuple of the zero or more child child hints
                        # subscripting this PEP 646-compliant unpacked child
                        # fixed-length tuple hint.
                        hint_child_childs = get_hint_pep_args(hint_child)
                        # print(f'Appending PEP 646 unpacked fixed-length tuple hint children {hint_child}...')

                        # Append all child child hints subscripting this PEP
                        # 646-compliant unpacked child fixed-length tuple hint
                        # to the end of this list, effectively unpacking this
                        # child hint directly into the new PEP 585-compliant
                        # parent fixed-length tuple hint to be returned.
                        hint_pep585_childs_list.extend(hint_child_childs)
                    # Else, this child hint is *NOT* a PEP 646-compliant
                    # unpacked child fixed-length tuple hint. But this child
                    # hint is PEP 646-compliant! By process of elimination, this
                    # child hint *MUST* be a PEP 646-compliant unpacked type
                    # variable tuple (e.g., the "*Ts" in "tuple[str, *Ts]"). In
                    # this case, notify logic below that this PEP 646-compliant
                    # parent tuple hint is irreducible to a PEP 585-compliant
                    # parent tuple hint.
                    else:
                        # print(f'Ignoring PEP 646 non-unpacked fixed-length tuple hint {hint_child}...')
                        hint_pep585_childs_list = None
                # Else, this iteration has already discovered a PEP
                # 646-compliant child hint of this parent tuple hint. Since the
                # currently visited child hint is also PEP 646-compliant, this
                # parent tuple hint is subscripted by two or more PEP
                # 646-compliant child hints and thus violates PEP 646. In this
                # case, raise an exception.
                else:
                    assert hint_child_pep646_index is not None
                    raise BeartypeDecorHintPep646Exception(  # pragma: no cover
                        f'{exception_prefix}PEP 646 tuple type hint {repr(hint)} '
                        f'erroneously subscripted by two (or more) '
                        f'PEP 646 unpacked child hints:\n'
                        f'* PEP 646 unpacked child hint {repr(hint_child_pep646)} '
                        f'at index {hint_child_pep646_index}.\n'
                        f'* PEP 646 unpacked child hint {repr(hint_child)} '
                        f'at index {hint_child_index}.'
                    )
            # Else, this child hint is PEP 646-noncompliant.
            #
            # If this PEP 646-compliant parent tuple hint is still possibly
            # reducible to a PEP 585-compliant parent tuple hint, append this
            # PEP 646-noncompliant child hint to the list of all such hints to
            # subscript this PEP 585-compliant parent tuple hint by.
            elif hint_pep585_childs_list is not None:
                # print(f'Appending non-PEP 646 child hint {hint_child}...')
                hint_pep585_childs_list.append(hint_child)
            # Else, this PEP 646-compliant parent tuple hint is no longer
            # reducible to a PEP 585-compliant parent tuple hint (presumably due
            # to being subscripted by one or more PEP 646-compliant unpacked
            # child hints). In this case, continue validating this hint to be
            # PEP 646-compliant.

        # If prior iteration decided this PEP 646-compliant parent tuple hint to
        # be reducible to a PEP 585-compliant parent tuple hint, coerce this
        # list into a tuple of child hints subscripting the latter.
        if hint_pep585_childs_list is not None:
            hint_pep585_childs = tuple(hint_pep585_childs_list)
        # Else, prior iteration decided this PEP 646-compliant parent tuple hint
        # to *NOT* be reducible to a PEP 585-compliant parent tuple hint.

    # ....................{ RETURN                         }....................
    # Reduce this non-trivial PEP 646-compliant tuple hint to either...
    hint_reduced = (
        #FIXME: *STOP DOING THIS* as soon as we implement a proper code
        #generator for PEP 646-compliant tuple hints, please. *sigh*
        # If this PEP 646-compliant parent tuple hint is irreducible to a PEP
        # 585-compliant parent tuple hint, the builtin "tuple" type as a
        # temporary means of shallowly ignoring *ALL* child hints subscripting
        # this hint. Although obviously non-ideal, this simplistic approach does
        # have the benefit of actually working -- an improvement over our prior
        # approach of raising fatal exceptions for these hints.
        tuple
        if hint_pep585_childs is None else
        # Else, this PEP 646-compliant parent tuple hint is reducible to a PEP
        # 585-compliant parent tuple hint. In this case, the latter.
        make_hint_pep484585_tuple_fixed(hint_pep585_childs)
    )

    # print(f'Reduced PEP 646 tuple hint {hint} to {hint_reduced}!')
    return hint_reduced



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **current package beartype import hook skippable module** (i.e.,
data module blacklisted by one or more import hooks published by the
:mod:`beartype.claw` subpackage and thus expected to *not* be type-checked by
the :func:`beartype.beartype` decorator).
'''

# ....................{ IMPORTS                            }....................
from beartype.typing import (
    List,
    Union,
)

# ....................{ PEP 526                            }....................
# Validate that *NO* import hooks installed by the caller apply to this
# submodule. In this case, assert that PEP 526-compliant annotated assignment
# statements are *NOT* appended with calls to beartype's statement-level
# beartype.door.die_if_unbearable() exception-raiser.

# Assert that a PEP 526-compliant assignment statement assigning an object
# violating the type hint annotating that statement raises *NO* exception.
in_most_familiar_cadence: str = b'In most familiar cadence, with the howl'
assert isinstance(in_most_familiar_cadence, bytes)

# ....................{ FUNCTIONS                          }....................
def with_the_howl(the_thunder_and_the_hiss: Union[str, complex]) -> (
    Union[complex, List[bytes]]):
    '''
    Arbitrary method neither implicitly *nor* explicitly type-checked by the
    :func:`beartype.beartype` decorator.
    '''

    # This means nothing to us. Nothing!
    return the_thunder_and_the_hiss

# Assert that a function call passed a parameter violating the type hint
# annotating that parameter raises *NO* exception.
of_homeless_streams = with_the_howl(
    b'The thunder and the hiss of homeless streams')
assert isinstance(of_homeless_streams, bytes)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **executable main beartype import hook non-executable submodule**
(i.e., data module *only* intended to be imported from other submodules of this
data package rather than directly executed as a script).
'''

# ....................{ CALLABLES                          }....................
def by_solemn_vision(and_bright_silver_dream: int) -> None:
    '''
    Arbitrary callable annotated by trivial PEP-compliant type hints.

    To enable the parent Python process to validate that this callable actually
    *was* successfully called, this callable prints the passed parameter as is
    to standard output.
    '''

    print(and_bright_silver_dream)
